#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2019- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""Python logging facilities use the built-in logging module.

Upon import, the gmxapi package sets a placeholder "NullHandler" to block
propagation of log messages to the root logger (and sys.stderr, if not handled).

If you want to see gmxapi logging output on `sys.stderr`, import `logging` in your
script or module and configure it. For the simplest case, consider
`logging.basicConfig`::

    >>> import logging
    >>> logging.basicConfig(level=logging.DEBUG)

For more advanced usage, consider attaching a
`logging.StreamHandler` to the ``gmxapi`` logger.

The gmxapi logging module adds an additional ``rank_tag`` log formatter field that can
be particularly helpful in ensemble MPI workflows.

Example::

    ch = logging.StreamHandler()
    # Optional: Set log level.
    ch.setLevel(logging.DEBUG)
    # Optional: create formatter and add to character stream handler
    formatter = logging.Formatter('%(levelname)s %(asctime)s:%(name)s %(rank_tag)s%(message)s')
    ch.setFormatter(formatter)
    # add handler to logger
    logging.getLogger('gmxapi').addHandler(ch)

To handle log messages that are issued while importing :py:mod:`gmxapi` and its submodules,
attach the handler before importing :py:mod:`gmxapi`

Each module in the gmxapi package uses its own hierarchical logger to allow
granular control of log handling (e.g. ``logging.getLogger('gmxapi.operation')``).
Refer to the Python :py:mod:`logging` module for information on connecting to and handling
logger output.
"""

__all__ = ["logger"]

# Import system facilities
import logging
from logging import getLogger, DEBUG, NullHandler

# Define `logger` attribute that is used by submodules to create sub-loggers.
logger = getLogger("gmxapi")
# Prevent gmxapi logs from reaching logging.lastResort (and printing to sys.stderr)
# if the user does not take action to handle logging.
logger.addHandler(NullHandler(level=DEBUG))

logger.info("Importing gmxapi.")



#!/usr/bin/env python
#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2020- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""Building block based Dockerfile generation for CI testing images.

Generates a set of docker images used for running GROMACS CI on Gitlab.
The images are prepared according to a selection of build configuration targets
that hope to cover a broad enough scope of different possible systems,
allowing us to check compiler types and versions, as well as libraries used
for accelerators and parallel communication systems. Each combinations is
described as an entry in the build_configs dictionary, with the script
analysing the logic and adding build stages as needed.

Based on the example script provided by the NVidia HPCCM repository.

Reference:
    `NVidia HPC Container Maker <https://github.com/NVIDIA/hpc-container-maker>`__

Authors:
    * Paul Bauer <paul.bauer.q@gmail.com>
    * Eric Irrgang <ericirrgang@gmail.com>
    * Joe Jordan <e.jjordan12@gmail.com>
    * Mark Abraham <mark.j.abraham@gmail.com>
    * Gaurav Garg <gaugarg@nvidia.com>

Usage::

    $ python3 scripted_gmx_docker_builds.py --help
    $ python3 scripted_gmx_docker_builds.py --format docker > Dockerfile && docker build .
    $ python3 scripted_gmx_docker_builds.py | docker build -

See Also:
    :file:`buildall.sh`

"""

import argparse
import collections
import collections.abc
import copy
import os
import packaging.version
import shlex
import typing

try:
    import utility
except ImportError:
    raise RuntimeError(
        "This module assumes availability of supporting modules in the same directory. Add the directory to "
        "PYTHONPATH or invoke Python from within the module directory so module location can be resolved."
    )


# Basic packages for all final images.
_common_packages = [
    "build-essential",
    "ca-certificates",
    "ccache",
    "cmake",
    "git",
    "gnupg",
    "gpg-agent",
    "less",
    "libfftw3-dev",
    "libhwloc-dev",
    "liblapack-dev",
    "libx11-dev",
    "moreutils",
    "ninja-build",
    "python3",
    "rsync",
    "valgrind",
    "vim",
    "wget",
    "xsltproc",
]

_opencl_extra_packages = [
    "nvidia-opencl-dev",
    # The following require apt_ppas=['ppa:intel-opencl/intel-opencl'] on Ubuntu prior to 22.04
    "intel-opencl-icd",
    "ocl-icd-libopencl1",
    "ocl-icd-opencl-dev",
    "opencl-headers",
]

_rocm_extra_packages = [
    # The following require
    #             apt_keys=['http://repo.radeon.com/rocm/rocm.gpg.key'],
    #             apt_repositories=['deb [arch=amd64] http://repo.radeon.com/rocm/apt/X.Y.Z/ ubuntu main']
    "clinfo",
    "libelf1",
]

_rocm_version_dependent_packages = [
    # The following require
    #             apt_keys=['http://repo.radeon.com/rocm/rocm.gpg.key'],
    #             apt_repositories=['deb [arch=amd64] http://repo.radeon.com/rocm/apt/X.Y.Z/ ubuntu main']
    "hipfft",
    "hipfft-dev",
    "rocfft",
    "rocfft-dev",
    "rocprim",
    "rocprim-dev",
    "rocm-dev",
    "rocm-opencl",
]

# Extra packages required to build CP2K
_cp2k_extra_packages = [
    "autoconf",
    "autogen",
    "automake",
    "autotools-dev",
    "bzip2",
    "less",
    "libtool",
    "make",
    "nano",
    "patch",
    "pkg-config",
    "python3",
    "unzip",
    "xxd",
    "zlib1g-dev",
    "libopenblas-dev",
]

# Extra packages needed to build Intel Compute Runtime
_intel_compute_runtime_extra_packages = [
    "intel-opencl-icd",
    "intel-level-zero-gpu",
    "level-zero",
    "libmfx1",
]

# Extra packages needed to build Python installations from source.
# For Ubuntu 22.04, 'python-*' is later replaced with 'python3-*'
_python_extra_packages = [
    "build-essential",
    "ca-certificates",
    "ccache",
    "curl",
    "git",
    "libbz2-dev",
    "libffi-dev",
    "liblzma-dev",
    "libncurses5-dev",
    "libncursesw5-dev",
    "libreadline-dev",
    "libsqlite3-dev",
    "libssl-dev",
    "llvm",
    "python-openssl",
    "vim",
    "wget",
    "zlib1g-dev",
]

# Extra packages needed for images for building documentation.
_docs_extra_packages = [
    "autoconf",
    "automake",
    "autopoint",
    "autotools-dev",
    "bison",
    "flex",
    "ghostscript",
    "graphviz",
    "help2man",
    "imagemagick",
    "libtool",
    "mscgen",
    "m4",
    "openssh-client",
    "plantuml",
    "texinfo",
    "texlive-latex-base",
    "texlive-latex-extra",
    "texlive-fonts-recommended",
    "texlive-fonts-extra",
    "tex-gyre",
]


def get_oneapi_repository(args) -> "hpccm.building_blocks.base":
    if args.ubuntu is None:
        raise RuntimeError("oneAPI only supported on Ubuntu")
    return hpccm.building_blocks.packages(
        apt_keys=[
            "https://apt.repos.intel.com/intel-gpg-keys/GPG-PUB-KEY-INTEL-SW-PRODUCTS.PUB"
        ],
        apt_repositories=[
            "deb [signed-by=/usr/share/keyrings/GPG-PUB-KEY-INTEL-SW-PRODUCTS.gpg arch=amd64] https://apt.repos.intel.com/oneapi all main"
        ],
    )


def get_impi_version(oneapi_version) -> str:
    # Map oneAPI version numbers to the version numbers of
    # Intel MPI that they contain.
    version_map = {
        "2025.3": "2021.17",
        "2025.2": "2021.16",
        "2025.1": "2021.15",
        "2025.0": "2021.14",
        "2024.2": "2021.13",
        "2024.1": "2021.12",
        "2024.0": "2021.11",
    }
    return version_map[oneapi_version]


# Parse command line arguments
parser = argparse.ArgumentParser(
    description="GROMACS CI image creation script", parents=[utility.parser]
)

parser.add_argument(
    "--format",
    type=str,
    default="docker",
    choices=["docker", "singularity"],
    help="Container specification format (default: %(default)s)",
)


def base_image_tag(args) -> str:
    """Generate *image* for hpccm.baseimage()."""
    # Check if we use CUDA images or plain linux images
    if args.cuda is not None:
        cuda_version_tag = "nvidia/cuda:" + args.cuda + "-devel"
        if args.centos is not None:
            cuda_version_tag += "-centos" + args.centos
        elif args.ubuntu is not None:
            cuda_version_tag += "-ubuntu" + args.ubuntu
        else:
            raise RuntimeError("Logic error: no Linux distribution selected.")

        base_image_tag = cuda_version_tag
    else:
        if args.centos is not None:
            base_image_tag = "centos:centos" + args.centos
        elif args.ubuntu is not None:
            base_image_tag = "ubuntu:" + args.ubuntu
        else:
            raise RuntimeError("Logic error: no Linux distribution selected.")
    return base_image_tag


def hpccm_distro_name(args) -> str:
    """Generate *_distro* for hpccm.baseimage().

    Convert the linux distribution variables into something that hpccm
    understands.

    The same format is used by the lower level hpccm.config.set_linux_distro().
    """
    if args.centos is not None:
        name_mapping = {"7": "centos7", "8": "centos8"}
        if args.centos in name_mapping:
            hpccm_name = name_mapping[args.centos]
        else:
            raise RuntimeError("Logic error: unsupported CentOS distribution selected.")
    elif args.ubuntu is not None:
        name_mapping = {
            "24.04": "ubuntu24",
            "22.04": "ubuntu22",
            "20.04": "ubuntu20",
            "18.04": "ubuntu18",
            "16.04": "ubuntu16",
        }
        if args.ubuntu in name_mapping:
            hpccm_name = name_mapping[args.ubuntu]
        else:
            raise RuntimeError("Logic error: unsupported Ubuntu distribution selected.")
    else:
        raise RuntimeError("Logic error: no Linux distribution selected.")
    return hpccm_name


def get_llvm_packages(args) -> typing.Iterable[str]:
    # If we use the package version of LLVM, we need to install extra packages for it.
    if (args.llvm is not None) and (args.tsan is None):
        packages = [
            f"libomp-{args.llvm}-dev",
            f"libomp5-{args.llvm}",
            f"libc++-{args.llvm}-dev",
            f"libc++1-{args.llvm}",
            f"libc++abi-{args.llvm}-dev",
            f"libc++abi1-{args.llvm}",
            "clang-format-" + str(args.llvm),
            "clang-tidy-" + str(args.llvm),
        ]
        if args.adaptivecpp is not None:
            packages += [
                f"llvm-{args.llvm}-dev",
                f"libclang-{args.llvm}-dev",
                f"libclang-rt-{args.llvm}-dev",
                f"lld-{args.llvm}",
            ]
        return packages
    else:
        return []


def get_opencl_packages(args) -> typing.List[str]:
    if (args.doxygen is None) and (args.oneapi is None):
        return _opencl_extra_packages
    else:
        return []


def get_rocm_packages(args) -> typing.List[str]:
    if args.rocm is None:
        return []
    else:
        packages = _rocm_extra_packages
        packages.extend(get_rocm_version_dependent_packages(args))
        return packages


def get_rocm_version_dependent_packages(args) -> typing.List[str]:
    packages = []
    rocm_version = args.rocm
    rocm_version_parsed = [int(i) for i in rocm_version.split(".")]
    # if the version does not contain the final patch version, add it manually to make sure that
    # we can install the packages
    if len(rocm_version_parsed) < 3:
        rocm_version = rocm_version + ".0"
    for entry in _rocm_version_dependent_packages:
        packages.append(entry + rocm_version)

    return packages


def get_rocm_repository(args) -> "hpccm.building_blocks.base":
    if args.ubuntu is None:
        raise RuntimeError("ROCm only supported on Ubuntu")
    else:
        rocm_version = [int(i) for i in args.rocm.split(".")]
        if rocm_version[0] < 5 or (rocm_version[0] == 5 and rocm_version[1] < 3):
            dist_string = "ubuntu"
        else:
            dist_string = {"20.04": "focal", "22.04": "jammy", "24.04": "noble"}[
                args.ubuntu
            ]
    return hpccm.building_blocks.packages(
        apt_keys=["http://repo.radeon.com/rocm/rocm.gpg.key"],
        apt_repositories=[
            f"deb [arch=amd64 signed-by=/usr/share/keyrings/rocm.gpg.gpg] http://repo.radeon.com/rocm/apt/{args.rocm}/ {dist_string} main"
        ],
    )


def get_cp2k_packages(args) -> typing.List[str]:
    cp2k_packages = []
    if args.cp2k:
        cp2k_packages.extend(_cp2k_extra_packages)
        if args.mpi is not None:
            cp2k_packages.append("libfftw3-mpi-dev")
    return cp2k_packages


def get_compiler(
    args, compiler_build_stage: "hpccm.Stage" = None
) -> "hpccm.building_blocks.base":
    # Compiler
    if args.llvm is not None:
        # Build our own version instead to get TSAN + OMP
        if args.tsan is not None:
            if compiler_build_stage is not None:
                compiler = compiler_build_stage.runtime(_from="tsan")
            else:
                raise RuntimeError("No TSAN compiler build stage!")
        # Build the default compiler if we don't need special support
        else:
            # Always use the "upstream" llvm repositories because the
            # main ubuntu repositories stop adding support for new
            # llvm versions after a few llvm releases.
            compiler = hpccm.building_blocks.llvm(version=args.llvm, upstream=True)

    elif args.oneapi is not None:
        if compiler_build_stage is not None:
            compiler = compiler_build_stage.runtime(_from="oneapi")
            # Prepare the toolchain (needed only for builds done within the Dockerfile, e.g.
            # OpenMPI builds, which don't currently work for other reasons)
            if args.oneapi.startswith("2024.0"):
                path = f"/opt/intel/oneapi/compiler/{args.oneapi}/linux/bin"
            else:
                path = f"/opt/intel/oneapi/compiler/{args.oneapi}/bin"
            oneapi_toolchain = hpccm.toolchain(
                CC=f"{path}/icx",
                CXX=f"{path}/icpx",
            )
            setattr(compiler, "toolchain", oneapi_toolchain)

        else:
            raise RuntimeError("No oneAPI compiler build stage!")

    elif args.intel_llvm is not None:
        if compiler_build_stage is not None:
            compiler = compiler_build_stage.runtime(_from="intel_llvm")
            intel_llvm_toolchain = hpccm.toolchain(
                CC="/opt/intel-llvm/bin/clang", CXX="/opt/intel-llvm/bin/clang++"
            )
            setattr(compiler, "toolchain", intel_llvm_toolchain)
        else:
            raise RuntimeError("No IntelLLVM compiler build stage!")

    elif args.gcc is not None:
        if args.cp2k is not None:
            compiler = hpccm.building_blocks.gnu(
                extra_repository=True, version=args.gcc, fortran=True
            )
        else:
            compiler = hpccm.building_blocks.gnu(
                extra_repository=True, version=args.gcc, fortran=False
            )
    else:
        raise RuntimeError("Logic error: no compiler toolchain selected.")
    return compiler


def get_gdrcopy(args, compiler):
    if args.cuda is not None:
        if hasattr(compiler, "toolchain"):
            # Version last updated August 15, 2024
            return hpccm.building_blocks.gdrcopy(
                toolchain=compiler.toolchain, version="2.4.1"
            )
        else:
            raise RuntimeError("compiler is not an HPCCM compiler building block!")
    else:
        return None


def get_ucx(args, compiler, gdrcopy):
    if args.cuda is not None or args.rocm is not None:
        if hasattr(compiler, "toolchain"):
            use_gdrcopy = gdrcopy is not None
            # We disable `-Werror`, since there are some unknown pragmas and unused variables which upset clang
            toolchain = copy.copy(compiler.toolchain)
            toolchain.CFLAGS = "-Wno-error"
            toolchain.CXXFLAGS = "-Wno-error"
            configure_opts = []
            if args.rocm is not None:
                configure_opts.append("--with-rocm=/opt/rocm")
            use_cuda = args.cuda is not None
            # Version last updated August 15, 2024
            return hpccm.building_blocks.ucx(
                toolchain=toolchain,
                gdrcopy=use_gdrcopy,
                version="1.19.1",
                cuda=use_cuda,
                configure_opts=configure_opts,
            )
        else:
            raise RuntimeError("compiler is not an HPCCM compiler building block!")
    else:
        return None


def get_mpi(args, compiler, ucx):
    # If needed, add MPI to the image
    if args.mpi is not None:
        if args.mpi == "openmpi":
            if hasattr(compiler, "toolchain"):
                if args.oneapi is not None:
                    raise RuntimeError("oneAPI building OpenMPI is not supported")
                use_cuda = args.cuda is not None
                configure_opts = []
                if args.rocm is not None:
                    configure_opts.append("--with-rocm=/opt/rocm")
                use_ucx = ucx is not None
                # Version last updated April 17, 2025
                return hpccm.building_blocks.openmpi(
                    toolchain=compiler.toolchain,
                    version="5.0.8",
                    cuda=use_cuda,
                    ucx=use_ucx,
                    infiniband=False,
                    configure_opts=configure_opts,
                )
            else:
                raise RuntimeError("compiler is not an HPCCM compiler building block!")
        elif args.mpi == "mpich":
            if hasattr(compiler, "toolchain"):
                use_cuda = args.cuda is not None
                use_rocm = args.rocm is not None
                use_ucx = ucx is not None
                flags = {}
                if ucx is not None:
                    flags["with-device"] = "ch4:ucx"
                mpich_stage = hpccm.Stage()
                # Python needed for configuring
                mpich_stage += hpccm.building_blocks.python(
                    python3=True, python2=False, devel=False
                )
                # Version last updated August 15, 2024
                mpich_stage += hpccm.building_blocks.mpich(
                    toolchain=compiler.toolchain,
                    version="4.2.2",
                    cuda=use_cuda,
                    rocm=use_rocm,
                    ucx=use_ucx,
                    infiniband=False,
                    disable_fortran=True,
                    **flags,
                )
                return mpich_stage
            else:
                raise RuntimeError("compiler is not an HPCCM compiler building block!")
        elif args.mpi == "impi":
            if args.oneapi is None:
                # Note IMPI works with all compilers, but GROMACS doesn't test that in CI
                raise RuntimeError("IMPI build requires oneAPI")
            impi_version = get_impi_version(args.oneapi)
            impi_stage = hpccm.Stage()
            impi_stage += get_oneapi_repository(args)
            impi_stage += hpccm.building_blocks.packages(
                # Add minimal packages (not the whole HPC toolkit!)
                ospackages=[
                    f"intel-oneapi-mpi-{impi_version}",
                    f"intel-oneapi-mpi-devel-{impi_version}",
                ],
            )
            return impi_stage
        else:
            raise RuntimeError("Requested unknown MPI implementation.")
    else:
        return None


def get_oneapi_plugins(args):
    blocks = []

    def _add_plugin(variant):
        if args.oneapi is None:
            raise RuntimeError("Cannot install oneAPI plugins without oneAPI.")
        backend_version = {"nvidia": args.cuda, "amd": args.rocm}[variant]
        if backend_version.count(".") == 2:
            backend_version = ".".join(backend_version.split(".")[:2])  # 12.0.1 -> 12.0
        oneapi_version = args.oneapi
        outfile = f"/tmp/oneapi_plugin_{variant}.sh"
        if packaging.version.Version(args.oneapi) >= packaging.version.Version(
            "2025.1"
        ):
            url = f"https://developer.codeplay.com/api/v1/products/download?product=oneapi&version={oneapi_version}&variant={variant}&filters[]=linux"
            blocks.append(
                hpccm.primitives.shell(
                    commands=[
                        f"wget --content-disposition '{url}' --output-document '{outfile}'",
                        # Work around unilateral assumption made in installer
                        f"mkdir -p /opt/intel/oneapi/{args.oneapi}/lib",
                        f"bash {outfile} --backend-version {backend_version} --yes",
                    ]
                )
            )
        else:
            url = f"https://developer.codeplay.com/api/v1/products/download?product=oneapi&version={oneapi_version}&variant={variant}&filters[]=linux&filters[]={backend_version}"
            blocks.append(
                hpccm.primitives.shell(
                    commands=[
                        f"wget --content-disposition '{url}' --output-document '{outfile}'",
                        f"bash '{outfile}' --yes",
                    ]
                )
            )

    if args.oneapi_plugin_nvidia:
        _add_plugin("nvidia")
    if args.oneapi_plugin_amd:
        _add_plugin("amd")
        # Need to make ROCm libraries discoverable
        blocks.append(
            hpccm.primitives.shell(
                commands=[
                    "echo '/opt/rocm/lib/' > /etc/ld.so.conf.d/rocm.conf",
                    "ldconfig",
                ]
            )
        )
    return blocks


def get_clfft(args):
    if args.clfft is not None:
        return hpccm.building_blocks.generic_cmake(
            repository="https://github.com/clMathLibraries/clFFT.git",
            prefix="/usr/local",
            recursive=True,
            branch=args.clfft,
            directory="clFFT/src",
        )
    else:
        return None


def get_heffte(args):
    if args.heffte is not None:
        opts = [
            "-D CMAKE_BUILD_TYPE=Release",
            "-D Heffte_ENABLE_FFTW=OFF",
            "-D BUILD_SHARED_LIBS=ON",
        ]
        if args.cuda is not None:
            opts.append("-D Heffte_ENABLE_CUDA=ON")
        if args.rocm is not None:
            opts.append("-D Heffte_ENABLE_ROCM=ON")
            opts.append("-D CMAKE_HIP_ARCHITECTURES=gfx1034,gfx906")

        return hpccm.building_blocks.generic_cmake(
            cmake_opts=opts,
            repository="https://github.com/icl-utk-edu/heffte",
            prefix="/usr/local",
            recursive=True,
            commit=args.heffte,
            directory="heffte",
        )
    else:
        return None


def get_plumed(args):
    if args.plumed is not None:
        return hpccm.building_blocks.generic_autotools(
            url="https://github.com/plumed/plumed2/releases/download/v2.9.2/plumed-src-2.9.2.tgz",
            prefix="/usr/local",
            runtime_environment={"PLUMED_KERNEL": "/usr/local/lib/libplumedKernel.so"},
            directory="plumed-2.9.2",
        )
    else:
        return None


def get_libtorch(args):
    if args.libtorch is not None:
        lt_version = args.libtorch
        cuda_version = args.cuda.replace(".", "")
        assert cuda_version[:-1] == "130", "Use CUDA 13.0 for libtorch"
        return hpccm.primitives.shell(
            commands=[
                "mkdir -p /var/tmp",
                f"wget -q -nc --no-check-certificate -O /var/tmp/libtorch.zip https://download.pytorch.org/libtorch/cu130/libtorch-shared-with-deps-{lt_version}%2Bcu130.zip",
                "unzip -q /var/tmp/libtorch.zip -d /usr/local/libtorch",
                "rm -f /var/tmp/libtorch.zip",
            ]
        )
    else:
        return None


def get_nvhpcsdk(args):
    if args.nvhpcsdk is not None:
        return hpccm.building_blocks.nvhpc(
            eula=True,
            cuda_multi=False,
            environment=False,
            mpi=False,
            version=args.nvhpcsdk,
        )
    else:
        return None


def get_adaptivecpp(args):
    if args.adaptivecpp is None:
        return None
    if args.rocm is None:
        raise RuntimeError("AdaptiveCpp requires the ROCm packages")
    if args.llvm is None:
        # We're using ROCm LLVM in this case, which is not compatible with CUDA
        if args.cuda is not None:
            raise RuntimeError(
                "Can not build AdaptiveCpp with CUDA and no upstream LLVM"
            )
    preconfigure = []

    if args.llvm is not None:
        cmake_opts = [
            "-DCMAKE_C_COMPILER=clang-{}".format(args.llvm),
            "-DCMAKE_CXX_COMPILER=clang++-{}".format(args.llvm),
            "-DLLVM_DIR=/usr/lib/llvm-{}/cmake/".format(args.llvm),
        ]
        acpp_major, acpp_minor, _ = map(int, args.adaptivecpp.split("."))
        if int(args.llvm) >= 18 and (
            acpp_major < 24 or acpp_major == 24 and acpp_minor <= 6
        ):
            preconfigure.append(
                "ln -s /usr/lib/llvm-{0}/lib/libLLVM-{0}.so /usr/lib/llvm-{0}/lib/libLLVM.so".format(
                    args.llvm
                )
            )
    else:
        cmake_opts = [
            "-DCMAKE_C_COMPILER=/opt/rocm/bin/amdclang",
            "-DCMAKE_CXX_COMPILER=/opt/rocm/bin/amdclang++",
            "-DLLVM_DIR=/opt/rocm/llvm/lib/cmake/llvm",
            "-DWITH_SSCP_COMPILER=OFF",
            "-DWITH_OPENCL_BACKEND=OFF",
            "-DWITH_LEVEL_ZERO_BACKEND=OFF",
        ]

    cmake_opts += [
        "-DCMAKE_PREFIX_PATH=/opt/rocm/lib/cmake",
        "-DWITH_ROCM_BACKEND=ON",
    ]

    if args.cuda is not None:
        cmake_opts += [
            "-DCUDAToolkit_ROOT=/usr/local/cuda",
            "-DWITH_CUDA_BACKEND=ON",
        ]

    return hpccm.building_blocks.generic_cmake(
        repository="https://github.com/AdaptiveCpp/AdaptiveCpp.git",
        branch=f"v{args.adaptivecpp}",
        directory="/var/tmp/AdaptiveCpp",
        prefix="/usr/local",
        recursive=True,
        preconfigure=preconfigure,
        cmake_opts=["-DCMAKE_BUILD_TYPE=Release", *cmake_opts],
    )


def get_cp2k(args):
    if args.cp2k is None:
        return None

    if args.gcc is None:
        raise RuntimeError("CP2K build requires GNU compilers")

    make_commands = ["make -j$(nproc) ARCH=local VERSION=ssmp libcp2k"]
    if args.mpi is not None:
        make_commands += ["make -j$(nproc) ARCH=local VERSION=psmp libcp2k"]
    make_commands += ["rm -rf ./obj"]

    return hpccm.building_blocks.generic_build(
        repository="https://github.com/cp2k/cp2k.git",
        branch=f"support/v{args.cp2k}",
        recursive=True,
        build=[
            "mkdir -p /opt/cp2k",
            "cp -rf ./* /opt/cp2k/",
            "cd /opt/cp2k/tools/toolchain",
            "./install_cp2k_toolchain.sh  \
                --with-gcc=system            \
                --with-cmake=system          \
                --with-openmpi=system        \
                --with-fftw=system           \
                --with-openblas=system       \
                --with-scalapack=install     \
                --with-gsl=no                \
                --with-elpa=no               \
                --with-spglib=no             \
                --with-spfft=no              \
                --with-cosma=no              \
                --with-libvori=no            \
                --with-sirius=no             \
                --with-hdf5=no               \
                --with-libxc=install         \
                --with-libxsmm=install       \
                --with-libint=install        \
                --libint-lmax=5",
            "rm -rf ./build",
            "cp ./install/arch/local.* ../../arch/",
            "cd ../../",
        ]
        + make_commands,
    )


def add_tsan_compiler_build_stage(
    input_args, output_stages: typing.Mapping[str, "hpccm.Stage"]
):
    """Isolate the expensive TSAN preparation stage.

    This is a very expensive stage, but has few and disjoint dependencies, and
    its output is easily compartmentalized (/usr/local) so we can isolate this
    build stage to maximize build cache hits and reduce rebuild time, bookkeeping,
    and final image size.
    """
    if not isinstance(output_stages, collections.abc.MutableMapping):
        raise RuntimeError("Need output_stages container.")
    if "compiler_build" in output_stages:
        raise RuntimeError('"compiler_build" output stage is already present.')
    tsan_stage = hpccm.Stage()
    tsan_stage += hpccm.primitives.baseimage(
        image=base_image_tag(input_args),
        _distro=hpccm_distro_name(input_args),
        _as="tsan",
    )

    tsan_stage += hpccm.building_blocks.packages(
        ospackages=["git", "ca-certificates", "build-essential", "cmake"]
    )
    # CMake will get duplicated later, but this is an expensive image, and it isn't worth optimizing
    # out that duplication...
    tsan_stage += hpccm.building_blocks.python(python3=True, python2=False, devel=False)

    compiler_branch = "release/" + str(input_args.llvm) + ".x"
    tsan_stage += hpccm.building_blocks.generic_cmake(
        repository="https://github.com/llvm/llvm-project.git",
        directory="/var/tmp/llvm-project/llvm/",
        prefix="/usr/local",
        recursive=True,
        branch=compiler_branch,
        cmake_opts=[
            "-D CMAKE_BUILD_TYPE=Release",
            '-D LLVM_ENABLE_PROJECTS="clang;openmp;clang-tools-extra;compiler-rt;lld"',
            "-D LIBOMP_TSAN_SUPPORT=on",
        ],
        postinstall=[
            "ln -s /usr/local/bin/clang++ /usr/local/bin/clang++-"
            + str(input_args.llvm),
            "ln -s /usr/local/bin/clang-format /usr/local/bin/clang-format-"
            + str(input_args.llvm),
            "ln -s /usr/local/bin/clang-tidy /usr/local/bin/clang-tidy-"
            + str(input_args.llvm),
            "ln -s /usr/local/share/clang/run-clang-tidy.py /usr/local/bin/run-clang-tidy-"
            + str(input_args.llvm)
            + ".py",
            "ln -s /usr/local/bin/run-clang-tidy-"
            + str(input_args.llvm)
            + ".py /usr/local/bin/run-clang-tidy-"
            + str(input_args.llvm),
            "ln -s /usr/local/libexec/c++-analyzer /usr/local/bin/c++-analyzer-"
            + str(input_args.llvm),
        ],
    )
    output_stages["compiler_build"] = tsan_stage


def oneapi_runtime(_from="0"):
    oneapi_runtime_stage = hpccm.Stage()
    oneapi_runtime_stage += hpccm.primitives.copy(
        _from="oneapi-build",
        files={"/opt/intel": "/opt/intel", "/etc/bash.bashrc": "/etc/bash.bashrc"},
    )
    return oneapi_runtime_stage


def add_oneapi_compiler_build_stage(
    input_args, output_stages: typing.Mapping[str, "hpccm.Stage"]
):
    """Isolate the oneAPI preparation stage.

    This stage is isolated so that its installed components are minimized in the
    final image (chiefly /opt/intel) and its environment setup script can be
    sourced. This also helps with rebuild time and final image size.
    """
    if not isinstance(output_stages, collections.abc.MutableMapping):
        raise RuntimeError("Need output_stages container.")
    if "compiler_build" in output_stages:
        raise RuntimeError('"compiler_build" output stage is already present.')
    oneapi_stage = hpccm.Stage()
    oneapi_stage += hpccm.primitives.baseimage(
        image=base_image_tag(input_args),
        _distro=hpccm_distro_name(input_args),
        _as="oneapi-build",
    )

    version = str(input_args.oneapi)

    # Add required components for the next stage (both for hpccm and Intel's setvars.sh script)
    oneapi_stage += hpccm.building_blocks.packages(
        ospackages=["wget", "gnupg2", "ca-certificates", "lsb-release"]
    )
    oneapi_stage += get_oneapi_repository(args)
    oneapi_stage += hpccm.building_blocks.packages(
        # Add minimal packages (not the whole HPC toolkit!)
        ospackages=[
            f"intel-oneapi-dpcpp-cpp-{version}",
            f"intel-oneapi-openmp-{version}",
            f"intel-oneapi-mkl-{version}",
            f"intel-oneapi-mkl-devel-{version}",
        ],
    )
    # Ensure that all bash shells on the final container will have access to oneAPI
    oneapi_stage += hpccm.primitives.shell(
        commands=[
            'echo "source /opt/intel/oneapi/setvars.sh" >> /etc/bash.bashrc',
            "unlink /opt/intel/oneapi/compiler/latest",
            f"ln -sf /opt/intel/oneapi/compiler/{version} /opt/intel/oneapi/compiler/latest",
        ]
    )
    setattr(oneapi_stage, "runtime", oneapi_runtime)

    output_stages["compiler_build"] = oneapi_stage


def intel_llvm_runtime(_from="0"):
    llvm_runtime_stage = hpccm.Stage()
    llvm_runtime_stage += hpccm.primitives.copy(
        _from="intel-llvm-build", files={"/opt/intel-llvm": "/opt/intel-llvm"}
    )

    bashrc = [
        "export DPCPP_HOME=/opt/intel-llvm",
        "export PATH=${DPCPP_HOME}/bin:$PATH",
        "export LD_LIBRARY_PATH=${DPCPP_HOME}/lib:${LD_LIBRARY_PATH}",
        'export CFLAGS="-isystem ${DPCPP_HOME}/include"',
        'export CXXFLAGS="-isystem ${DPCPP_HOME}/include"',
    ]
    # Since we cannot just create a file, we write to it line-by-line using "echo".
    # We must shlex.quote the lines to ensure all spaces/quotes/etc are preserved.
    commands = [
        "echo {} >> /opt/intel-llvm/setenv.sh".format(shlex.quote(line))
        for line in bashrc
    ] + ['echo source "/opt/intel-llvm/setenv.sh" >> /etc/bash.bashrc']
    llvm_runtime_stage += hpccm.primitives.shell(commands=commands)

    return llvm_runtime_stage


def add_intel_llvm_compiler_build_stage(
    input_args, output_stages: typing.Mapping[str, "hpccm.Stage"]
):
    """Isolate the Intel LLVM (open-source oneAPI) preparation stage.

    This stage is isolated so that its installed components are minimized in the
    final image (chiefly /opt/intel) and its environment setup script can be
    sourced. This also helps with rebuild time and final image size.
    """
    if not isinstance(output_stages, collections.abc.MutableMapping):
        raise RuntimeError("Need output_stages container.")
    if "compiler_build" in output_stages:
        raise RuntimeError('"compiler_build" output stage is already present.')
    llvm_stage = hpccm.Stage()
    llvm_stage += hpccm.primitives.baseimage(
        image=base_image_tag(input_args),
        _distro=hpccm_distro_name(input_args),
        _as="intel-llvm-build",
    )
    buildbot_flags = [
        "--build-type=Release",
        "--cuda",  # Build with CUDA support
        "--llvm-external-projects=openmp",  # Enable OpenMP
        "--obj-dir=/var/tmp/llvm/llvm/build",  # Build directory
        # Disable OpenMP offload targets we're not using
        "--cmake-opt=-DLIBOMPTARGET_BUILD_AMDGPU_PLUGIN=FALSE",
        "--cmake-opt=-DLIBOMPTARGET_BUILD_CUDA_PLUGIN=FALSE",
        "--cmake-opt=-DLIBOMPTARGET_BUILD_DEVICERTL_BCLIB=FALSE",
        # Help CMake find CUDA Driver stub, see https://github.com/opencv/opencv/issues/6577
        "--cmake-opt=-DCMAKE_LIBRARY_PATH=/usr/local/cuda/targets/x86_64-linux/lib/stubs/",
    ]

    llvm_stage += hpccm.building_blocks.packages(
        ospackages=[
            "git",
            "ninja-build",
            "cmake",
            "python3",
            "python3-dev",
            "build-essential",
            "wget",
        ]
    )

    if args.rocm is not None:
        llvm_stage += get_rocm_repository(args)
        llvm_stage += hpccm.building_blocks.packages(
            ospackages=get_rocm_packages(args), aptitude=True
        )
        buildbot_flags.extend(["--hip", "--hip-platform", "AMD"])

    llvm_stage += hpccm.building_blocks.generic_build(
        repository="https://github.com/intel/llvm.git",
        directory="llvm/llvm",
        build=[
            "mkdir -p /var/tmp/llvm/llvm/build",
            shlex.join(
                ["python3", "/var/tmp/llvm/buildbot/configure.py", *buildbot_flags]
            ),
            "cd /var/tmp/llvm/llvm/build",
            # Must be called after the configure.py
            shlex.join(
                [
                    "cmake",
                    "/var/tmp/llvm/llvm",
                    "-DCMAKE_INSTALL_PREFIX=/opt/intel-llvm/",
                ]
            ),
            "ninja",
            "ninja sycl-toolchain install",
        ],
        install=[],
        branch=input_args.intel_llvm,
    )

    setattr(llvm_stage, "runtime", intel_llvm_runtime)

    output_stages["compiler_build"] = llvm_stage


def prepare_venv(version: packaging.version.Version) -> typing.Sequence[str]:
    """Get shell commands to set up the venv for the requested Python version."""
    major = version.major
    minor = version.minor  # type: int

    pyenv = "$HOME/.pyenv/bin/pyenv"

    py_ver = f"{major}.{minor}"
    venv_path = f"$HOME/venv/py{py_ver}"
    commands = [
        f"$({pyenv} prefix `{pyenv} whence python{py_ver}`)/bin/python -m venv {venv_path}"
    ]

    commands.append(f"{venv_path}/bin/python -m pip install --upgrade pip setuptools")
    # Install dependencies for building and testing gmxapi Python package.
    # WARNING: Please keep this list synchronized with python_packaging/gmxapi/requirements.txt
    #  and docs/requirements.txt
    # TODO: Get requirements.txt from an input argument.
    commands.append(
        f"""{venv_path}/bin/python -m pip install --upgrade \
            'black' \
            'breathe' \
            'build' \
            'cmake>=3.28' \
            'flake8>=3.7.7' \
            'furo' \
            'gcovr>=4.2' \
            'importlib-resources;python_version<"3.10"' \
            'mypy' \
            'networkx>=2.0' \
            'numpy>1.7' \
            'packaging' \
            'pip>=10.1' \
            'pybind11>=2.12' \
            'Pygments>=2.2.0' \
            'pytest>=4.6' \
            'python-gitlab' \
            'setuptools>=61' \
            'Sphinx>=4.0' \
            'sphinx-argparse' \
            'sphinx-copybutton' \
            'sphinx_inline_tabs' \
            'sphinxcontrib-autoprogram' \
            'sphinxcontrib-plantuml>=0.14' \
            'versioningit>=2' \
            'wheel'"""
    )
    return commands


def get_cmake_stages(*, input_args: argparse.Namespace, base: str):
    """Get the stage(s) necessary for the requested CMake versions.

    One (intermediate) build stage is created
    for each CMake version, based on the *base* stage.
    See ``--cmake`` option.

    Each stage uses the version number to determine an installation location:
        /usr/local/cmake-{version}

    The resulting path is easily copied into the main stage.

    Returns:
        dict of isolated CMake installation stages with keys from ``cmake-{version}``
    """
    cmake_stages = {}
    for cmake_version in input_args.cmake:
        stage_name = f"cmake-{cmake_version}"
        cmake_stages[stage_name] = hpccm.Stage()
        cmake_stages[stage_name] += hpccm.primitives.baseimage(
            image=base, _distro=hpccm_distro_name(input_args), _as=stage_name
        )
        cmake_stages[stage_name] += hpccm.building_blocks.cmake(
            eula=True, prefix=f"/usr/local/{stage_name}", version=cmake_version
        )
    return cmake_stages


def add_python_stages(
    input_args: argparse.Namespace,
    *,
    base: str,
    output_stages: typing.MutableMapping[str, "hpccm.Stage"],
):
    """Add the stage(s) necessary for the requested venvs.

    One intermediate build stage is created for each venv (see --venv option).

    Each stage partially populates Python installations and venvs in the home
    directory. The home directory is collected by the 'pyenv' stage for use by
    the main build stage.
    """
    if len(input_args.venvs) < 1:
        raise RuntimeError("No venvs to build...")
    if output_stages is None or not isinstance(output_stages, collections.abc.Mapping):
        raise RuntimeError("Need a container for output stages.")

    # Main Python stage that collects the environments from individual stages.
    # We collect the stages individually, rather than chaining them, because the
    # copy is a bit slow and wastes local Docker image space for each filesystem
    # layer.
    pyenv_stage = hpccm.Stage()
    pyenv_stage += hpccm.primitives.baseimage(
        image=base, _distro=hpccm_distro_name(input_args), _as="pyenv"
    )
    python_extra_packages = _python_extra_packages
    if input_args.ubuntu is not None and input_args.ubuntu != "20.04":
        python_extra_packages = [
            i.replace("python-", "python3-") for i in python_extra_packages
        ]
    pyenv_stage += hpccm.building_blocks.packages(ospackages=python_extra_packages)

    for version in [
        packaging.version.parse(py_ver) for py_ver in sorted(input_args.venvs)
    ]:
        stage_name = "py" + str(version)
        stage = hpccm.Stage()
        stage += hpccm.primitives.baseimage(
            image=base, _distro=hpccm_distro_name(input_args), _as=stage_name
        )
        stage += hpccm.building_blocks.packages(ospackages=python_extra_packages)

        # TODO: Use a non-root user for testing and Python virtual environments.
        stage += hpccm.primitives.shell(
            commands=[
                "curl https://pyenv.run | bash",
                """echo 'export PYENV_ROOT="$HOME/.pyenv"' >> $HOME/.bashrc""",
                """echo 'export PATH="$PYENV_ROOT/bin:$PATH"' >> $HOME/.bashrc""",
                """echo 'eval "$(pyenv init -)"' >> $HOME/.bashrc""",
                """echo 'eval "$(pyenv virtualenv-init -)"' >> $HOME/.bashrc""",
            ]
        )
        pyenv = "$HOME/.pyenv/bin/pyenv"
        commands = [
            f'PYTHON_CONFIGURE_OPTS="--enable-shared" {pyenv} install -s {version}'
        ]
        stage += hpccm.primitives.shell(commands=commands)

        commands = prepare_venv(version)
        stage += hpccm.primitives.shell(commands=commands)

        # TODO: Update user home directory.
        pyenv_stage += hpccm.primitives.copy(
            _from=stage_name, _mkdir=True, src=["/root/"], dest="/root"
        )

        # Add the intermediate build stage to the sequence
        output_stages[stage_name] = stage

    # TODO: If we activate pyenv for login shells, the `global` "version" should be full-featured.
    # # `version` should be a system installation or pyenv environment (or pyenv-virtualenv)
    # # with the dependencies for all of the Python aspects of CMake-driven builds.
    # commands = '{pyenv} global {version}'.format(
    #             pyenv=pyenv,
    #             version=...)
    # pyenv_stage += hpccm.primitives.shell(commands=commands)

    # Add the aggregating build stage to the sequence. This allows the main stage to copy
    # the files in a single stage, potentially reducing the overall output image size.
    output_stages["pyenv"] = pyenv_stage


def add_documentation_dependencies(
    input_args, output_stages: typing.MutableMapping[str, "hpccm.Stage"]
):
    """Add appropriate layers according to doxygen input arguments."""
    if input_args.doxygen is None:
        return
    # In recent python, it's a warning to install via pip into the
    # system environment, so we use pipx (as recommended).
    output_stages["main"] += hpccm.building_blocks.packages(ospackages=["pipx"])
    # Always clone the same version of linkchecker (latest release at June 1, 2021)
    output_stages["main"] += hpccm.primitives.shell(
        commands=[
            # Default linkchecker detects when it is run as root and
            # unilaterally de-elevates privileges, which means it
            # can't search its own installation path for files it
            # needs. Since we are using an ephemeral container, it's
            # fine to run as root, so we use a version of 10.0.1
            # hacked to avoid the call to drop_privileges().
            "pipx install git+https://github.com/mabraham/linkchecker.git@avoid-drop-privileges"
        ]
    )
    # Add the path to which pipx installs binaries to the PATH
    output_stages["main"] += hpccm.primitives.environment(
        variables={"PATH": "${PATH}:/root/.local/bin"}
    )
    output_stages["main"] += hpccm.primitives.shell(
        commands=[
            'sed -i \'/"XPS"/d;/"PDF"/d;/"PS"/d;/"EPS"/d;/disable ghostscript format types/d\' /etc/ImageMagick-6/policy.xml'
        ]
    )
    if input_args.doxygen == "1.8.5":
        doxygen_commit = "ed4ed873ab0e7f15116e2052119a6729d4589f7a"
        output_stages["main"] += hpccm.building_blocks.generic_autotools(
            repository="https://github.com/westes/flex.git",
            commit="f7788a9a0ecccdc953ed12043ccb59ca25714018",
            prefix="/tmp/install-of-flex",
            configure_opts=["--disable-shared"],
            preconfigure=["./autogen.sh"],
        )
        output_stages["main"] += hpccm.building_blocks.generic_autotools(
            repository="https://github.com/doxygen/doxygen.git",
            commit=doxygen_commit,
            prefix="",
            configure_opts=["--flex /tmp/install-of-flex/bin/flex", "--static"],
        )
    else:
        version = input_args.doxygen
        archive_name = f"doxygen-{version}.linux.bin.tar.gz"
        archive_url = f"https://sourceforge.net/projects/doxygen/files/rel-{version}/{archive_name}"
        binary_path = f"doxygen-{version}/bin/doxygen"
        commands = [
            "mkdir doxygen && cd doxygen",
            f"wget {archive_url}",
            f"tar xf {archive_name} {binary_path}",
            f"cp {binary_path} /usr/local/bin/",
            "cd .. && rm -rf doxygen",
        ]
        output_stages["main"] += hpccm.primitives.shell(commands=commands)


def add_base_stage(
    name: str, input_args, output_stages: typing.MutableMapping[str, "hpccm.Stage"]
):
    """Establish dependencies that are shared by multiple parallel stages."""
    # Building blocks are chunks of container-builder instructions that can be
    # copied to any build stage with the addition operator.
    building_blocks = collections.OrderedDict()
    building_blocks["base_packages"] = hpccm.building_blocks.packages(
        ospackages=_common_packages
    )

    # These are the most expensive and most reusable layers, so we put them first.
    building_blocks["compiler"] = get_compiler(
        input_args, compiler_build_stage=output_stages.get("compiler_build")
    )
    if args.rocm is not None:
        building_blocks["rocm"] = [
            get_rocm_repository(args),
            hpccm.building_blocks.packages(ospackages=get_rocm_packages(args)),
        ]
    building_blocks["gdrcopy"] = get_gdrcopy(input_args, building_blocks["compiler"])
    building_blocks["ucx"] = get_ucx(
        input_args, building_blocks["compiler"], building_blocks["gdrcopy"]
    )
    building_blocks["mpi"] = get_mpi(
        input_args, building_blocks["compiler"], building_blocks["ucx"]
    )

    # Create the stage from which the targeted image will be tagged.
    output_stages[name] = hpccm.Stage()

    output_stages[name] += hpccm.primitives.baseimage(
        image=base_image_tag(input_args),
        _distro=hpccm_distro_name(input_args),
        _as=name,
    )
    for bb in building_blocks.values():
        if bb is not None:
            output_stages[name] += bb


def get_cross_compilation_packages(args):
    """Get packages needed for cross-compilation."""
    if args.cross is None:
        return []

    arch_triplet = f"{args.cross}-linux-gnu"
    packages = [
        "qemu-user",
        "qemu-user-static",
        "binfmt-support",
        "libtool",
        f"crossbuild-essential-{args.cross}",
        f"gcc-{args.gcc}-{arch_triplet}",
        f"g++-{args.gcc}-{arch_triplet}",
    ]
    return packages


def add_cross_compilation_support(
    args, output_stages: typing.MutableMapping[str, "hpccm.Stage"]
):
    """Configure cross-compilation toolchain and QEMU support."""
    if args.cross is None:
        return

    arch_triplet = f"{args.cross}-linux-gnu"

    # Determine compiler settings based on which compiler is specified
    if args.llvm is not None:
        c_compiler = f"clang-{args.llvm}"
        cxx_compiler = f"clang++-{args.llvm}"
        # Add additional compiler flags needed for cross-compiling with LLVM
        compiler_flags = f"--target={arch_triplet}"
    else:
        # Default to GCC
        c_compiler = f"{arch_triplet}-gcc-{args.gcc}"
        cxx_compiler = f"{arch_triplet}-g++-{args.gcc}"
        compiler_flags = ""
        qemu_flags = ""

    if args.cross == "riscv64" and args.llvm is not None:
        # QEmu support VLen up to 1024 bits
        compiler_flags += " -march=rv64gcv_zvl1024b"
        qemu_flags += " -cpu rv64,v=true,vext_spec=v1.0,vlen=1024"

    # Add host architecture environment for CMake cross-compilation
    output_stages["main"] += hpccm.primitives.shell(
        commands=[
            # Create a CMake toolchain file for cross-compilation
            f"mkdir -p /opt/cross",
            f"echo 'set(CMAKE_SYSTEM_NAME Linux)' > /opt/cross/{args.cross}-toolchain.cmake",
            f"echo 'set(CMAKE_SYSTEM_PROCESSOR {args.cross})' >> /opt/cross/{args.cross}-toolchain.cmake",
            f"echo 'set(CMAKE_C_COMPILER {c_compiler})' >> /opt/cross/{args.cross}-toolchain.cmake",
            f"echo 'set(CMAKE_CXX_COMPILER {cxx_compiler})' >> /opt/cross/{args.cross}-toolchain.cmake",
            f"echo 'set(CMAKE_C_FLAGS_INIT \"{compiler_flags}\")' >> /opt/cross/{args.cross}-toolchain.cmake",
            f"echo 'set(CMAKE_CXX_FLAGS_INIT \"{compiler_flags}\")' >> /opt/cross/{args.cross}-toolchain.cmake",
            f"echo 'set(CMAKE_FIND_ROOT_PATH /usr/{arch_triplet})' >> /opt/cross/{args.cross}-toolchain.cmake",
            f"echo 'set(CMAKE_FIND_ROOT_PATH_MODE_PROGRAM NEVER)' >> /opt/cross/{args.cross}-toolchain.cmake",
            f"echo 'set(CMAKE_FIND_ROOT_PATH_MODE_LIBRARY ONLY)' >> /opt/cross/{args.cross}-toolchain.cmake",
            f"echo 'set(CMAKE_FIND_ROOT_PATH_MODE_INCLUDE ONLY)' >> /opt/cross/{args.cross}-toolchain.cmake",
            f"echo 'set(CMAKE_CROSSCOMPILING_EMULATOR /usr/local/bin/run-{args.cross})' >> /opt/cross/{args.cross}-toolchain.cmake",
            # Configure QEMU for this architecture
            f"echo '#!/bin/bash' > /usr/local/bin/run-{args.cross}",
            f"echo 'export QEMU_LD_PREFIX=/usr/{arch_triplet}' >> /usr/local/bin/run-{args.cross}",
            f"echo 'qemu-{args.cross} {qemu_flags} \"$@\"' >> /usr/local/bin/run-{args.cross}",
            f"chmod +x /usr/local/bin/run-{args.cross}",
        ]
    )

    # Set environment variables for cross-compilation
    output_stages["main"] += hpccm.primitives.environment(variables=env_vars)


def build_stages(args) -> typing.Iterable["hpccm.Stage"]:
    """Define and sequence the stages for the recipe corresponding to *args*."""

    # A Dockerfile or Singularity recipe can have multiple build stages.
    # The main build stage can copy files from previous stages, though only
    # the last stage is included in the tagged output image. This means that
    # large or expensive sets of build instructions can be isolated in
    # local/temporary images, but all of the stages need to be output by this
    # script, and need to occur in the correct order, so we create a sequence
    # object early in this function.
    stages = collections.OrderedDict()

    # If we need TSAN or oneAPI support the early build is more complex,
    # so that our compiler images don't have all the cruft needed to get those things
    # installed.
    if args.llvm is not None and args.tsan is not None:
        add_tsan_compiler_build_stage(input_args=args, output_stages=stages)
    if args.oneapi is not None:
        add_oneapi_compiler_build_stage(input_args=args, output_stages=stages)
    if args.intel_llvm is not None:
        add_intel_llvm_compiler_build_stage(input_args=args, output_stages=stages)

    add_base_stage(name="build_base", input_args=args, output_stages=stages)

    # Add Python environments to MPI images, only, so we don't have to worry
    # about whether to install mpi4py.
    if args.mpi is not None and len(args.venvs) > 0:
        add_python_stages(base="build_base", input_args=args, output_stages=stages)

    # Building blocks are chunks of container-builder instructions that can be
    # copied to any build stage with the addition operator.
    building_blocks = collections.OrderedDict()

    # Install additional packages early in the build to optimize Docker build layer cache.
    os_packages = (
        list(get_llvm_packages(args))
        + get_opencl_packages(args)
        + get_cp2k_packages(args)
        + get_cross_compilation_packages(args)
    )
    if args.doxygen is not None:
        os_packages += _docs_extra_packages
    if args.oneapi is not None:
        os_packages += ["lsb-release"]
    if args.adaptivecpp is not None:
        acpp_major, acpp_minor, _ = map(int, args.adaptivecpp.split("."))
        if acpp_major < 25 or (acpp_major == 25 and acpp_minor < 10):
            os_packages += ["libboost-fiber-dev"]
    if args.hdf5:
        os_packages += ["libhdf5-dev"]
    if args.libtorch is not None:
        os_packages += ["unzip"]
    building_blocks["extra_packages"] = []
    if args.intel_compute_runtime:
        if args.ubuntu == "24.04":
            # Per https://dgpu-docs.intel.com/driver/client/overview.html#ubuntu-latest
            building_blocks["extra_packages"] += hpccm.building_blocks.packages(
                apt_ppas=["ppa:kobuk-team/intel-graphics"]
            )
            os_packages += [
                "libze-intel-gpu1",
                "libze1",
                "intel-metrics-discovery",
                "intel-opencl-icd",
                "clinfo",
                "intel-gsc",
            ]

        else:
            repo = {
                "22.04": "deb [signed-by=/usr/share/keyrings/intel-graphics.gpg arch=amd64] https://repositories.intel.com/gpu/ubuntu jammy client",
                "20.04": "deb [signed-by=/usr/share/keyrings/intel-graphics.gpg arch=amd64] https://repositories.intel.com/graphics/ubuntu focal main",
            }
            building_blocks["extra_packages"] += hpccm.building_blocks.packages(
                apt_keys=["https://repositories.intel.com/gpu/intel-graphics.key"],
                apt_repositories=[repo[args.ubuntu]],
            )
            os_packages += _intel_compute_runtime_extra_packages

    if args.ubuntu is not None and args.ubuntu == "20.04":
        building_blocks["extra_packages"] += hpccm.building_blocks.packages(
            ospackages=os_packages, apt_ppas=["ppa:intel-opencl/intel-opencl"]
        )
    else:
        # For 22.04, everything is packaged. Therefore, we don't need Intel repo, but we need to
        # use aptitude to resolve the conflicts between Ubuntu ROCm packages and AMD ROCm packages.
        building_blocks["extra_packages"] += hpccm.building_blocks.packages(
            ospackages=os_packages, aptitude=True
        )

    building_blocks["CP2K"] = get_cp2k(args)

    if args.cuda is not None and args.llvm is not None:
        # Hack to tell clang what version of CUDA we're using
        # based on https://github.com/llvm/llvm-project/blob/1fdec59bffc11ae37eb51a1b9869f0696bfd5312/clang/lib/Driver/ToolChains/Cuda.cpp#L43
        cuda_version_split = args.cuda.split(".")
        # LLVM requires having the version in x.y.z format, while args.cuda be be either x.y or x.y.z
        cuda_version_str = "{}.{}.{}".format(
            cuda_version_split[0],
            cuda_version_split[1],
            cuda_version_split[2] if len(cuda_version_split) > 2 else 0,
        )
        building_blocks["cuda-clang-workaround"] = hpccm.primitives.shell(
            commands=[
                f'echo "CUDA Version {cuda_version_str}" > /usr/local/cuda/version.txt'
            ]
        )

    building_blocks["oneapi_plugins"] = get_oneapi_plugins(args)

    building_blocks["clfft"] = get_clfft(args)

    building_blocks["heffte"] = get_heffte(args)

    building_blocks["plumed"] = get_plumed(args)

    building_blocks["libtorch"] = get_libtorch(args)

    building_blocks["nvhpcsdk"] = get_nvhpcsdk(args)
    if building_blocks["nvhpcsdk"] is not None:
        nvshmem_lib_path = (
            "/opt/nvidia/hpc_sdk/Linux_x86_64/"
            + args.nvhpcsdk
            + "/comm_libs/nvshmem/lib/:$LD_LIBRARY_PATH"
        )
        building_blocks["nvhpcsdk_path"] = hpccm.primitives.environment(
            variables={"LD_LIBRARY_PATH": nvshmem_lib_path}
        )

    building_blocks["AdaptiveCpp"] = get_adaptivecpp(args)

    # Add Python environments to MPI images, only, so we don't have to worry
    # about whether to install mpi4py.
    if (args.mpi is not None) and len(args.venvs) > 0:
        add_python_stages(base="build_base", input_args=args, output_stages=stages)

    cmake_stages = get_cmake_stages(input_args=args, base="build_base")
    stages.update(cmake_stages)

    # Create the stage from which the targeted image will be tagged.
    stages["main"] = hpccm.Stage()

    stages["main"] += hpccm.primitives.baseimage(
        image="build_base", _distro=hpccm_distro_name(args), _as="main"
    )
    for bb in building_blocks.values():
        if bb is not None:
            stages["main"] += bb

    # We always add Python3 and Pip
    stages["main"] += hpccm.building_blocks.python(
        python3=True, python2=False, devel=False if args.libtorch is None else True
    )

    # Add documentation requirements (doxygen and sphinx + misc).
    if args.doxygen is not None:
        add_documentation_dependencies(args, stages)

    for stage_name in cmake_stages:
        stages["main"] += hpccm.primitives.copy(
            _from=stage_name,
            _mkdir=True,
            src=[f"/usr/local/{stage_name}/"],
            dest=f"/usr/local/{stage_name}",
        )

    if "pyenv" in stages and stages["pyenv"] is not None:
        stages["main"] += hpccm.primitives.copy(
            _from="pyenv", _mkdir=True, src=["/root/.pyenv/"], dest="/root/.pyenv"
        )
        stages["main"] += hpccm.primitives.copy(
            _from="pyenv", _mkdir=True, src=["/root/venv/"], dest="/root/venv"
        )
        # TODO: Update user home directory.
        # TODO: If we activate pyenv for login shells, the `global` "version" should be full-featured.
        # stages['main'] += hpccm.primitives.copy(_from='pyenv', src=['/root/.bashrc'],
        #                                         dest='/root/')

    # Make sure that `python` resolves to something.
    stages["main"] += hpccm.primitives.shell(
        commands=[
            "test -x /usr/bin/python || "
            "update-alternatives --install /usr/bin/python python /usr/bin/python3 1 && "
            "/usr/bin/python --version"
        ]
    )

    # Add cross-compilation support if requested
    if args.cross is not None:
        add_cross_compilation_support(args, stages)

    # Note that the list of stages should be sorted in dependency order.
    for build_stage in stages.values():
        if build_stage is not None:
            yield build_stage


if __name__ == "__main__":
    args = parser.parse_args()
    import hpccm.config

    # Set container specification output format
    hpccm.config.set_container_format(args.format)
    # Normally in hpccm the first call to baseimage sets the context
    # for other packages, e.g. which distro version to use. We want to
    # set that early on, so that hpccm can do the right thing to use
    # the right llvm apt repo when we want to use versions of llvm
    # that were never supported by the main apt repo.
    hpccm.config.set_linux_distro(hpccm_distro_name(args))

    container_recipe = build_stages(args)

    # Output container specification
    for stage in container_recipe:
        print(stage)



#!/usr/bin/env python3
#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2022- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""Script to dispatch a post-merge-acceptance pipeline.

Submits a new post-merge pipeline for the specified branch, useful for
testing issues that only arise in those jobs.

Requires the `python_gitlab <https://python-gitlab.readthedocs.io/en/stable/index.html>`__
package for Python access to the GitLab API.

Author:
    * Paul Bauer <paul.bauer.q@gmail.com>
"""

import argparse
import warnings

try:
    import gitlab
except ImportError:
    warnings.warn(
        "This tool requires the `gitlab` package. Try `pip install python-gitlab`."
    )
    gitlab = None


def submit_gitlab_pipeline(
    auth_token,
    pipeline_type="POST_MERGE_ACCEPTANCE",
    branch="main",
    regtest_branch="main",
    regtest_commit="FETCH_HEAD",
):
    """Submit a post merge pipeline to GitLab server.

    The pipeline is by default created for main, but can be run for any branch.

    Needs access to gitlab python API. For this, someone running the script needs
    to have installed the python library (pip install python-gitlab) and created
    a personal access token that grants access to the GROMACS project
    (https://docs.gitlab.com/ee/user/profile/personal_access_tokens.html).

    Throws if the authentication process, or the submission of the pipeline fails.
    """
    gl = gitlab.Gitlab("https://gitlab.com", private_token=auth_token)
    # The project ID for GROMACS is hardcoded here
    project_id = 17679574
    project = gl.projects.get(project_id)

    print("Going to start pipeline with following arguments")
    print(r"Run for = ", branch)
    print(r"Type = ", pipeline_type)
    print(r"Regression tests from: ", regtest_branch)
    print(r"Regressiontests commit: ", regtest_commit)

    return project.pipelines.create(
        {
            "ref": branch,
            "variables": [
                {"key": pipeline_type, "value": "true"},
                {"key": "REGRESSIONTESTBRANCH", "value": regtest_branch},
                {"key": "REGRESSIONTESTCOMMIT", "value": regtest_commit},
            ],
        }
    )


parser = argparse.ArgumentParser(
    description="Options for manually submitting pipelines."
)

parser.add_argument(
    "--type",
    type=str,
    default="POST_MERGE_ACCEPTANCE",
    help='What kind of pipeline to run (default is "POST_MERGE_ACCEPTANCE")',
)

parser.add_argument(
    "--token",
    type=str,
    required=True,
    help="GitLab access token needed to launch pipelines",
)

parser.add_argument(
    "--branch",
    type=str,
    default="main",
    help='Branch to run pipeline for (default "main")',
)

parser.add_argument(
    "--regtest-branch",
    type=str,
    default="",
    help="Regressiontest branch to use to for running regression tests (default none, which means fall back to main)",
)

parser.add_argument(
    "--regtest-commit",
    type=str,
    default="",
    help="Commit to use instead of the regtest-branch tip for running tests (default empty)",
)

if __name__ == "__main__":
    args = parser.parse_args()
    submit_gitlab_pipeline(
        args.token, args.type, args.branch, args.regtest_branch, args.regtest_commit
    )



#!/usr/bin/env python
# This script takes an ipython notebook as an argument and rewrites the file
# with metadata stripped to make change tracking with git easier.
import sys
import json
import os
import shutil

infile = sys.argv[1]
if not os.path.exists(infile):
    sys.exit("command line argument must be an existing filename")
tempfile = infile + ".tmp"

with open(infile, "r") as fh:
    json_in = json.load(fh)

nb_metadata = json_in["metadata"]


def strip_output_from_cell(cell):
    if "outputs" in cell:
        cell["outputs"] = []
    if "execution_count" in cell:
        cell["execution_count"] = None


for cell in json_in["cells"]:
    strip_output_from_cell(cell)

with open(tempfile, "w") as fh:
    json.dump(json_in, fh, sort_keys=True, indent=1, separators=(",", ": "))
shutil.move(tempfile, infile)



###########################################################################
#                                                                         #
#    physical_validation,                                                 #
#    a python package to test the physical validity of MD results         #
#                                                                         #
#    Written by Michael R. Shirts <michael.shirts@colorado.edu>           #
#               Pascal T. Merz <pascal.merz@colorado.edu>                 #
#                                                                         #
#    Copyright (C) 2012 University of Virginia                            #
#              (C) 2017 University of Colorado Boulder                    #
#                                                                         #
#    This library is free software; you can redistribute it and/or        #
#    modify it under the terms of the GNU Lesser General Public           #
#    License as published by the Free Software Foundation; either         #
#    version 2.1 of the License, or (at your option) any later version.   #
#                                                                         #
#    This library is distributed in the hope that it will be useful,      #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of       #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    #
#    Lesser General Public License for more details.                      #
#                                                                         #
#    You should have received a copy of the GNU Lesser General Public     #
#    License along with this library; if not, write to the                #
#    Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,     #
#    Boston, MA 02110-1301 USA                                            #
#                                                                         #
###########################################################################
r"""
This module contains low-level functionality of the
`physical_validation.integrator` module. The functions in this module should
generally not be called directly. Please use the high-level functions from
`physical_validation.integrator`.
"""
import numpy as np

from . import plot


def calculate_rmsd(data, time=None, slope=False):
    assert isinstance(data, np.ndarray) and data.ndim == 1
    assert time is None or isinstance(time, np.ndarray) and time.ndim == 1

    avg = data.mean()

    if time is None:
        time = np.arange(data.size)

    fit = np.polyfit(time, data, 1)

    def f(x): return fit[0]*x + fit[1]

    if slope:
        rmsd = 0
        for t, d in zip(time, data):
            rmsd += (d - f(t))**2
        rmsd = np.sqrt(rmsd / data.size)
    else:
        rmsd = data.std()

    return avg, rmsd, fit[0]


def max_deviation(dts, rmsds):
    dt_ratio_2 = (dts[:-1] / dts[1:])**2
    rmsds = rmsds[:-1] / rmsds[1:]
    return np.max(np.abs(1 - rmsds/dt_ratio_2))


def check_convergence(const_traj,
                      convergence_test=max_deviation,
                      verbose=True, slope=False,
                      screen=False, filename=None):

    assert isinstance(const_traj, dict)
    assert len(const_traj) >= 2

    if verbose:
        print('{:65s}'.format('-'*65))
        print('{:>10s} {:>10s} {:>10s} {:>10s} {:^21s}'.format('dt', 'avg', 'rmsd', 'slope', 'ratio'))
        print('{:43s} {:>10s} {:>10s}'.format('', 'dt^2', 'rmsd'))
        print('{:65s}'.format('-'*65))

    prev = None

    results = {}
    for dt, traj in sorted(const_traj.items(), key=lambda x: float(x[0]), reverse=True):
        assert isinstance(traj, np.ndarray)
        assert traj.ndim == 1 or traj.ndim == 2
        dt = float(dt)

        if traj.ndim == 1:
            data = traj
            time = None
        else:
            data = traj[1]
            time = traj[0]

        results[dt] = calculate_rmsd(data, time, slope)

        if verbose:
            if prev is None:
                print('{:10.4g} {:10.2f} {:10.2e} {:10.2e} {:>10s} {:>10s}'.format(dt, results[dt][0], results[dt][1],
                                                                                   results[dt][2], '--', '--'))
                prev = [dt, results[dt][1]]
            else:
                print('{:10.4g} {:10.2f} {:10.2e} {:10.2e} {:10.2f} {:10.2f}'.format(dt, results[dt][0], results[dt][1],
                                                                                     results[dt][2],
                                                                                     prev[0]**2/dt**2,
                                                                                     prev[1]/results[dt][1]))
                prev = [dt, results[dt][1]]

    if verbose:
        print('{:65s}'.format('-'*65))

    dts = np.sort(np.array([float(dt) for dt in results.keys()]))[::-1]
    rmsds = np.array([float(results[dt][1]) for dt in dts])

    do_plot = screen or filename is not None

    if do_plot:
        data = [{'x': dts[1:],
                 'y': rmsds[:-1] / rmsds[1:],
                 'name': 'Integrator convergence'},
                {'x': dts[1:],
                 'y': (dts[:-1] / dts[1:])**2,
                 'name': 'Expected convergence'}]

        plot.plot(data,
                  legend='best',
                  title='Actual vs. expected convergence',
                  xlabel='Time step',
                  ylabel='Convergence',
                  xlim=(0, dts[1]),
                  inv_x=True,
                  filename=filename,
                  screen=screen)

    return convergence_test(dts, rmsds)



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2019- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

# This file is based on the Kasson Lab gmxapi project release 0.0.7.4.
# https://github.com/kassonlab/gmxapi/blob/v0.0.7.4/src/gmx/workflow.py
# # https://github.com/kassonlab/gmxapi/blob/v0.0.7.4/LICENSE
"""
Provide workflow-level utilities and classes
============================================

Supports the implementation of operations in the gmxapi.simulation module.
"""

__all__ = ["from_tpr", "WorkSpec", "WorkElement"]

import collections
import warnings
import weakref
from typing import Text, Iterable, Set

import gmxapi as gmx
from gmxapi import exceptions

# Module-level logger
logger = gmx.logger.getChild("simulation.workflow")
logger.info("Importing gmx.workflow")

# Work specification version string.
workspec_version = "gmxapi_workspec_0_1"
logger.info("Using schema version {}.".format(workspec_version))


def to_utf8(input) -> bytes:
    """Return a utf8 encoded byte sequence of the Unicode ``input`` or its string representation.

    Returns:
         :py:bytes byte sequence.
    """
    if isinstance(input, str):
        value = input.encode("utf-8")
    elif isinstance(input, bytes):
        value = input
    else:
        try:
            string = str(input)
            value = string.encode("utf-8")
        except Exception as e:
            raise exceptions.ValueError(
                "Input cannot be interpreted as a UTF-8 compatible string."
            ) from e
    return value


def to_string(input) -> str:
    """Return a Unicode string representation of ``input``.

    If ``input`` or its string representation is not already a Unicode object, attempt to decode as utf-8.

    Returns a native string, decoding utf-8 encoded byte sequences if necessary.
    """
    if isinstance(input, str):
        value = input
    else:
        try:
            value = input.decode("utf-8")
        except Exception:
            try:
                value = str(input)
            except Exception as e:
                raise exceptions.ValueError(
                    "Cannot find a string representation of input."
                ) from e
    return value


class GmxMap(dict):
    """Utility/compatibility class to ensure consistent keys.

    Internally, converts all keys to native str for the current interpreter.
    """

    def keys(self):
        for key in dict.keys(self):
            if not isinstance(key, str):
                raise exceptions.ApiError(
                    "Invalid key type found: {} {}".format(key, type(key))
                )
            yield key

    def __getitem__(self, key):
        return super(GmxMap, self).__getitem__(str(key))

    def __setitem__(self, key, item):
        super(GmxMap, self).__setitem__(str(key), item)

    def __delitem__(self, key):
        super(GmxMap, self).__delitem__(str(key))


class WorkSpec(object):
    """
    Container of workflow elements with data dependency
    information and requirements for execution.

    An element cannot be added to a WorkSpec if it has dependencies that are not
    in the WorkSpec.

    Work is added to the specification by passing a WorkElement object to
    :py:func:`WorkSpec.add_element()`.
    Any dependencies in the WorkElement must already be specified in the target WorkSpec.

    When iterated over, a WorkSpec object returns WorkElement objects.
    WorkElement objects are yielded in a valid order to keep dependencies
    satisfied, but not necessarily the same order in which add_element()
    calls were originally made. In other words, the WorkSpec is a directed
    acyclic dependency graph, and its iterator returns nodes in an arbitrary
    but topologically correct order.

    The string representation of a WorkSpec object is a valid JSON serialized data object.

    The schema for version 0.1 of the specification allows data structures like
    the following.
    ::

        {
            'version': 'gmxapi_workspec_0_1',
            'elements':
            {
                'myinput':
                {
                    'namespace': 'gromacs',
                    'operation': 'load_tpr',
                    'params': {'input': ['tpr_filename1', 'tpr_filename2']}
                },
                'mydata':
                {
                    'namespace': 'gmxapi',
                    'operation': 'open_global_data_with_barrier',
                    'params': ['data_filename']
                },
                'mypotential':
                {
                    'namespace': 'myplugin',
                    'operation': 'create_mdmodule',
                    'params': {...},
                    'depends': ['mydata']
                },
                'mysim':
                {
                    'namespace': 'gmxapi',
                    'operation': 'md',
                    'depends': ['myinput', 'mypotential']
                }
            }
        }

    The first mapping (``version``) is required as shown. The ``elements`` map
    contains uniquely named elements specifying an operation, the operation's
    namespace, and parameters and dependencies of the operation for this element.
    ``depends`` is a sequence of string names of elements that are also in the
    work spec. ``params`` is a key-value map with string keys and values that
    are valid JSON data. ``namespace`` and ``operation`` are strings that the
    :py:class:`Context <gmx.context.Context>` can map to directors it uses to
    construct the session. Namespace ``gmxapi`` is reserved for operations
    specified by the API. Namespace ``gromacs`` is reserved for operations
    implemented as GROMACS adapters (versioned separately from gmxapi). The
    period character (".") has special meaning and should not be used in naming
    elements, namespaces, or operations.

    """

    def __init__(self):
        self.version = workspec_version
        self.elements = GmxMap()
        self.__context_weak_ref = None

    @property
    def _context(self):
        referent = None
        if self.__context_weak_ref is not None:
            referent = self.__context_weak_ref()
        return referent

    @_context.setter
    def _context(self, context):
        # We're moving towards having the context own the work, so the work should
        # not own the context.
        self.__context_weak_ref = weakref.ref(context)

    def __chase_deps(self, source_set: Set[str], name_list: Iterable[Text]):
        """Helper to recursively generate dependencies before dependents.

        Given a set of WorkElement objects and a list of element names, generate WorkElements for
        the members of name_list plus their dependencies in an order such that dependencies are
        guaranteed to occur before their dependent elements.

        For example, to sequence an entire work specification into a reasonable order for instantiation, use

            >>> workspec.__chase_deps(set(workspec.elements.keys()), list(workspec.elements.keys()))

        Note: as a member function of WorkSpec, we have access to the full WorkSpec elements data at all
        times, giving us extra flexibility in implementation and arguments.

        Args:
            source_set: a (super)set of element names from the current work spec (will be consumed)
            name_list: subset of *sources* to be sequenced

        Returns:
            Sequence of WorkElement objects drawn from the names in *source_set*

        Requires that WorkElements named in *name_list* and any elements on which
        they depend are all named in *source_list* and available in the current
        work spec.

        Warning: *source_set* is a reference to an object that is modified arbitrarily.
        The caller should not re-use the object after calling _chase_deps().
        (Make a copy first, if needed.)

        TODO: Separate out DAG topology operations from here and Context.__enter__()
        Our needs are simple enough that we probably don't need an external dependency
        like networkx...
        """
        # Recursively (depth-first) generate a topologically valid serialized DAG from source_set.
        assert isinstance(source_set, set)
        if isinstance(name_list, (str, bytes)):
            warnings.warn(
                "name_list appears to be a single name. Disambiguate a string by passing a list or tuple."
            )
        assert isinstance(name_list, collections.abc.Iterable)

        # Make a copy of name_list in case the input reference is being used elsewhere during
        # iteration, such as for source_set, which is modified during the loop.
        for name in tuple(name_list):
            assert isinstance(name, str)
            if name in source_set:
                source_set.remove(name)
                element = WorkElement.deserialize(
                    self.elements[name], name=name, workspec=self
                )
                dependencies = element.depends
                # items in element.depends are either element names or ensembles of element names.
                for item in dependencies:
                    if isinstance(item, (list, tuple, set)):
                        dependency_list = item
                    else:
                        if not isinstance(item, str):
                            raise exceptions.ValueError(
                                "Dependencies should be a string or sequence of strings. Got {}".format(
                                    type(item)
                                )
                            )
                        dependency_list = [item]
                    for dependency in dependency_list:
                        for recursive_dep in self.__chase_deps(
                            source_set, (dependency,)
                        ):
                            yield recursive_dep
                yield element
            else:
                # Note: The user is responsible for ensuring that source_set is complete.
                # Otherwise, we would need to maintain a list of elements previously yielded.
                pass

    def __iter__(self):
        source_set = set(self.elements.keys())
        for element in self.__chase_deps(source_set, source_set):
            yield element

    def __hash__(self):
        """Uniquely identify this work specification.

        Allows the spec to be used as a dictionary key in Python. Note that this hash is possibly dependent on the
        Python implementation. It is not part of the gmxapi specification and should not be used outside of a single
        invocation of a script.
        """
        # Hash the serialized elements, concatenated as a single string. Note that the order of elements and their
        # contents is not guaranteed, but should be consistent within a script invocation.
        return hash(to_string(self.serialize()))

    def add_element(self, element):
        """Add an element to a work specification if possible.

        Adding an element to a WorkSpec must preserve the validity of the workspec, which involves several checks.
        We do not yet check for element uniqueness beyond a string name.

        If an element is added that was previously in another WorkSpec, it must first be removed from the
        other WorkSpec.
        """
        if (
            hasattr(element, "namespace")
            and hasattr(element, "operation")
            and hasattr(element, "serialize")
        ):
            if (
                not hasattr(element, "name")
                or element.name is None
                or len(str(element.name)) < 1
            ):
                raise exceptions.UsageError(
                    "Only named elements may be added to a WorkSpec."
                )
            if element.name in self.elements:
                raise exceptions.UsageError(
                    "Elements in WorkSpec must be uniquely identifiable."
                )
            if hasattr(element, "depends"):
                for dependency in element.depends:
                    if not dependency in self.elements:
                        raise exceptions.UsageError(
                            "Element dependencies must already be specified before an Element may be added."
                        )
            # Okay, it looks like we have an element we can add
            if (
                hasattr(element, "workspec")
                and element.workspec is not None
                and element.workspec is not self
            ):
                raise exceptions.Error(
                    "Element must be removed from its current WorkSpec to be added to this WorkSpec, but element "
                    "removal is not yet implemented."
                )
            self.elements[element.name] = element.serialize()
            element.workspec = self
        else:
            raise exceptions.ValueError(
                "Provided object does not appear to be compatible with gmx.workflow.WorkElement."
            )
        logger.info("Added element {} to workspec.".format(element.name))

    def serialize(self):
        """Serialize the work specification in a form suitable to pass to any Context implementation.

        Serialization is performed with the JSON data serialization module.

        To simplify unique identification of work specifications, this function will also impose rules for reproducibility.

        1. All key-value maps are sorted alphanumerically by their string keys.
        2. Strings must consist of valid ASCII characters.
        3. Output is a byte sequence of the utf-8 encoded densely formatted JSON document.

        Returns:
            ``unicode`` object in Python 2, ``bytes`` object in Python 3

        Output of serialize() should be explicitly converted to a string before passing to a JSON deserializer.

            >>> my_object = my_workspec.serialize()
            >>> my_data_structure = json.loads(my_object.decode('utf-8'))
            >>> # or...
            >>> my_data_structure = json.loads(my_object, encoding='utf-8')

        """
        import json

        # Build the normalized dictionary
        dict_representation = {"version": self.version, "elements": {}}
        for name, element in [
            (e, json.loads(to_string(self.elements[e])))
            for e in sorted(self.elements.keys())
        ]:
            dict_representation["elements"][str(name)] = element
        serialization = json.dumps(
            dict_representation,
            ensure_ascii=True,
            sort_keys=True,
            separators=(",", ":"),
        )
        return serialization.encode("utf-8")

    @classmethod
    def deserialize(serialized):
        import json

        workspec = WorkSpec()
        dict_representation = json.loads(to_string(serialized))
        ver_in = dict_representation["version"]
        ver_out = workspec.version
        if ver_in != ver_out:
            message = "Expected work spec version {}. Got work spec version {}.".format(
                ver_out, ver_in
            )
            raise exceptions.ValueError(message)
        for element in dict_representation["elements"]:
            workspec.elements[element] = dict_representation["elements"][element]
        return workspec

    def uid(self):
        """Get a unique identifier for this work specification.

        Returns:
            hash value

        Generate a cryptographic hash of this work specification that is guaranteed to match that of another equivalent
        work specification. The returned string is a 64-character hexadecimal encoded SHA-256 hash digest of the
        serialized WorkSpec.

        The definition of equivalence is likely to evolve, but currently means a work spec of the
        same version with the same named elements containing the same operations, dependencies, and parameters, as
        represented in the serialized version of the work specification. Note that this does not include checks on the
        actual contents of input files or anything that does not appear in the work specification directly. Also, the
        hash is lossy, so it is remotely conceivable that two specs could have the same hash. The work specs
        should be compared before making any expensive decisions based on work spec equivalence, such as with hash(workspec).

        Element names probably shouldn't be included in the unique identifying information (so that we can optimize out
        duplicated artifacts), but they are. A future API specification may add unique identification to the elements...
        """
        # Get an alphanumeric string of the checksum of the serialized work spec. SHA-256 should require about 43 characters
        # of base64 to represent, which seems reasonable. We need to replace some of the base64 characters to make them
        # filesystem friendly, though. Hexadecimal may be more friendly, but would require 64 characters.
        import hashlib

        data = to_utf8(self.serialize())
        result = hashlib.sha256(data)
        return result.hexdigest()

    def __str__(self):
        """Generate string representation for str() or print().

        The string output should look like the abstract schema for gmxapi_workspec_1_0, but the exact
        format is unspecified and may change in future versions.

        For consistent JSON output, use WorkSpec.serialize().
        """
        import json

        string = to_string(self.serialize())
        data = json.loads(string)
        reserialized = json.dumps(data, indent=4, sort_keys=True)
        return str(reserialized)

    def __repr__(self):
        """Generate Pythonic representation for repr(workspec)."""
        return "gmx.workflow.WorkSpec()"


# A possible alternative name for WorkElement would be Operator, since there is a one-to-one
# mapping between WorkElements and applications of "operation"s. We need to keep in mind the
# sensible distinction between the WorkElement abstraction and the API objects and DAG nodes.
class WorkElement(object):
    """Encapsulate an element of a work specification."""

    def __init__(self, namespace="gmxapi", operation=None, params=None, depends=()):
        self._namespace = str(to_string(namespace))
        # We can add an operations submodule to validate these. E.g. self.operation = gmx.workflow.operations.normalize(operation)
        if operation is not None:
            self._operation = str(to_string(operation))
        else:
            raise exceptions.UsageError("Invalid argument type for operation.")

        # Note: Nothing currently prevents attribute updates by assignment after adding the element to a workspec,
        # but this protocol will be clarified with https://github.com/kassonlab/gmxapi/issues/92
        if params is None:
            self.params = GmxMap()
        elif isinstance(params, dict):
            self.params = GmxMap({to_string(name): params[name] for name in params})
        else:
            raise exceptions.UsageError(
                "If provided, params must be a dictionary of keyword arguments"
            )
        self.depends = []
        for d in depends:
            if isinstance(d, (list, tuple)):
                self.depends.append([str(name) for name in d])
            else:
                self.depends.append(str(d))

        # The Python class for work elements keeps a strong reference to a WorkSpec object containing its description
        self._name = None
        self._workspec = None

    @property
    def namespace(self):
        assert isinstance(self._namespace, str)
        return self._namespace

    @property
    def operation(self):
        assert isinstance(self._operation, str)
        return self._operation

    @property
    def name(self):
        assert isinstance(self._name, (str, type(None)))
        return self._name

    @name.setter
    def name(self, new_name):
        self._name = str(to_string(new_name))

    @property
    def workspec(self):
        return self._workspec

    @workspec.setter
    def workspec(self, input):
        self._workspec = input

    def add_dependency(self, element):
        """Add another element as a dependency.

        First move the provided element to the same WorkSpec, if not already here.
        Then, add to ``depends`` and update the WorkSpec.
        """

        def check_element(element):
            if element.workspec is None:
                self.workspec.add_element(element)
                assert element.workspec is self.workspec
                assert element.name in self.workspec.elements
            elif element.workspec is not self.workspec:
                raise exceptions.ApiError(
                    "Element will need to be moved to the same workspec."
                )
            return True

        if hasattr(element, "workspec") and hasattr(element, "name"):
            check_element(element)
            self.depends.append(element.name)
        else:
            assert isinstance(element, (list, tuple))
            self.depends.append(
                tuple([item.name for item in element if check_element(item)])
            )

        self.workspec.elements[self.name] = self.serialize()

    def serialize(self):
        """Create a byte sequence representation of the work element.

        The WorkElement class exists just to provide convenient handles in Python. The WorkSpec is not actually a
        container of WorkElement objects.

        Returns:
            Byte sequence of utf-8 encoded JSON document. May need to be decoded if needed as a (Unicode) string.

        """
        import json

        output_dict = {
            "namespace": self.namespace,
            "operation": self.operation,
            "params": self.params,
            "depends": self.depends,
        }
        serialization = json.dumps(output_dict, default=lambda x: x.to_list())
        return to_utf8(serialization)

    @classmethod
    def deserialize(cls, input, name=None, workspec=None):
        """Create a new WorkElement object from a serialized representation.

        Arguments:
            input: a serialized WorkElement
            name: new element name (optional) (deprecated)
            workspec: an existing workspec to attach this element to (optional)

        When subclasses become distinct, this factory function will need to do additional dispatching to create an object of the correct type.
        Alternatively, instead of subclassing, a slightly heavier single class may suffice, or more flexible duck typing might be better.
        """
        import json

        input_string = to_string(input)
        args = json.loads(input_string)
        element = cls(
            namespace=args["namespace"],
            operation=args["operation"],
            params=args["params"],
            depends=args["depends"],
        )
        if name is not None:
            element.name = name
            # This conditional is nested because we can only add named elements to a WorkSpec.
            if workspec is not None:
                element.workspec = workspec
                if element.name not in workspec.elements:
                    workspec.add_element(element)
        return element


class SharedDataElement(WorkElement):
    """Work element with MD-specific extensions.

    The schema may not need to be changed, but the API object may be expected to provide additional functionality.
    """

    def __init__(self, params, name=None):
        """Create a blank SharedDataElement representation.

        It may be appropriate to insist on creating objects of this type via helpers or factories, particularly if
        creation requires additional parameters.
        """
        self.args = params["args"]
        self.kwargs = params["kwargs"]
        super(SharedDataElement, self).__init__(
            namespace="gmxapi",
            operation="global_data",
            params={"args": self.args, "kwargs": self.kwargs},
        )
        self.name = name


def get_source_elements(workspec):
    """Get an iterator of the starting nodes in the work spec.

    Source elements have no dependencies and can be processed immediately. Elements with dependencies
    cannot be processed, instantiated, or added to a work spec until after their dependencies have been.

    Args:
        workspec : an existing work specification to analyze, such as by a Context implementation preparing to schedule work.

    Returns:
        iterator of gmx.workflow.WorkElement objects that may be processed without dependencies.

    This function is provided in the API to allow flexibility in how source elements are determined.
    """
    for name in workspec.elements:
        element_data = workspec.elements[name]
        element = WorkElement.deserialize(element_data)
        if len(element.depends) == 0:
            element.name = name
            element.workspec = workspec
            yield (element)


def from_tpr(input=None, **kwargs):
    """Create a WorkSpec from a (list of) tpr file(s).

    Generates a work specification based on the provided simulation input and returns a handle to the
    MD simulation element of the workflow. Key word arguments can override simulation behavior from
    ``input``.

    If the MD operation discovers artifacts from a previous simulation that was launched from the same input,
    the simulation resumes from the last checkpointed step. If ``append_output`` is set ``False``, existing
    artifacts are kept separate from new output with the standard file naming convention,
    and new output begins from the last checkpointed step, if any.

    Setting ``end_time`` redefines the end point of the simulation trajectory from what was provided in
    ``input``. It is equivalent to changing the number of steps requested in the MDP (or TPR) input, but
    the time is provided as picoseconds instead of a number of time steps.

    .. deprecated:: 0.0.7
        If ``steps=N`` is provided and N is an integer
        greater than or equal to 1, the MD operation advances the trajectory by ``N`` steps, regardless of the number
        of simulation steps specified in ``input`` or ``end_time``. For convenience, setting ``steps=None`` does not override
        ``input``.
        Note that when it is not ``None``, ``steps`` takes precedence over ``end_time`` and ``input``, but can still be
        superceded by a signal, such as if an MD plugin or other code has a simulation completion condition that occurs
        before ``N`` additional steps have run.

    Where key word arguments correspond to ``gmx mdrun`` command line options, the corresponding flags are noted below.

    Keyword Arguments:
        input (str): *Required* string or list of strings giving the filename(s) of simulation input
        append_output (bool): Append output for continuous trajectories if True, truncate existing output data if False. (default True)
        end_time (float): Specify the final time in the simulation trajectory, overriding input read from TPR.
        grid (tuple): Domain decomposition grid divisions (nx, ny, nz). (-dd)
        max_hours (float): Terminate after 0.99 times this many hours if simulation is still running. (-maxh)
        pme_ranks (int): number of separate ranks to be used for PME electrostatics. (-npme)
        threads_per_pme_rank (int): Number of OpenMP threads per PME rank. (-ntomp_pme)
        steps (int): Override input files and run for this many steps. (-nsteps; deprecated)
        threads (int): Total number of threads to start for thread-MPI GROMACS. (-nt)
        threads_per_rank (int): number of OpenMP threads to start per MPI rank. (-ntomp)
        tmpi (int): number of thread-MPI ranks to start. (-ntmpi)

    ..  versionchanged:: 0.1
        *pme_threads_per_rank* renamed to *threads_per_pme_rank*.

    Note that run time simulator arguments can also be provided through `gmxapi.mdrun` *runtime_args*.

    Returns:
        simulation member of a gmx.workflow.WorkSpec object

    Produces a WorkSpec with the following data::

        version: gmxapi_workspec_0_1
        elements:
            tpr_input:
                namespace: gromacs
                operation: load_tpr
                params: {'input': ['tpr_filename1', 'tpr_filename2', ...]}
            md_sim:
                namespace: gmxapi
                operation: md
                depends: ['tpr_input']
                params: {'kw1': arg1, 'kw2': arg2, ...}

    Bugs: version 0.0.6
        * There is not a way to programatically check the current step number on disk.
          See https://github.com/kassonlab/gmxapi/issues/56 and https://github.com/kassonlab/gmxapi/issues/85

    """
    import os
    from gmxapi.utility import config

    usage = "argument to from_tpr() should be a valid filename or list of filenames, followed by optional key word arguments."

    # Normalize to tuple input type.
    if isinstance(input, list) or isinstance(input, tuple):
        tpr_list = tuple([to_string(element) for element in input])
    else:
        try:
            tpr_list = (to_string(input),)
        except:
            raise exceptions.UsageError(usage)

    # Check for valid filenames
    for arg in tpr_list:
        if not (os.path.exists(arg) and os.path.isfile(arg)):
            arg_path = os.path.abspath(arg)
            raise exceptions.UsageError(usage + " Got {}".format(arg_path))

    # Prepare a message about the build configuration.
    # From gmxapi/CMakeLists.txt, this is "library", "tmpi", or None
    mpi_type = config().get("gmx_mpi_type")
    mpi_message = ""
    if not mpi_type:
        mpi_message = "Currently using GROMACS that was built with no MPI."
    elif mpi_type == "tmpi":
        mpi_message = "Currently using GROMACS that was built with internal thread-MPI."
    elif mpi_type == "library":
        mpi_message = (
            "Currently using GROMACS that was built against an external MPI library."
        )
    else:
        warnings.warn(
            f"The gmxapi Python package does not recognize the GROMACS library MPI type: {mpi_type}. "
            "Compatibility checks may be impaired."
        )

    # Note: These are runner parameters, not MD parameters, and should be in the call to gmx.run() instead of here.
    # Reference https://github.com/kassonlab/gmxapi/issues/95
    # Also note that it is not well defined whether or which arguments should be allowed to differ within
    # an ensemble. But to allow parallel trajectory continuation, we have to accept distinct `-cpi` arguments.
    # Note that this violates the traditional gmxapi assumption that all ranks see the same
    # instructions. The "md_sim" element of the gmxapi 0.0.7 workspec ends up being unique
    # to the rank that sees it.
    params = {}
    if "-noappend" in kwargs:
        kwargs["append_output"] = False
        del kwargs["-noappend"]
    for arg_key in kwargs:
        if arg_key == "grid" or arg_key == "dd":
            params["grid"] = tuple(kwargs[arg_key])
        elif arg_key == "pme_ranks" or arg_key == "npme":
            params["pme_ranks"] = int(kwargs[arg_key])
        elif arg_key == "threads" or arg_key == "nt":
            if mpi_type == "tmpi":
                params["threads"] = int(kwargs[arg_key])
            else:
                raise exceptions.ValueError(
                    f"'{arg_key}' argument requires thread-MPI GROMACS. " + mpi_message
                )
        elif arg_key == "tmpi" or arg_key == "ntmpi":
            if mpi_type == "tmpi":
                params["tmpi"] = int(kwargs[arg_key])
            else:
                raise exceptions.ValueError(
                    f"'{arg_key}' argument requires thread-MPI GROMACS. " + mpi_message
                )
        elif arg_key == "threads_per_rank" or arg_key == "ntomp":
            params["threads_per_rank"] = int(kwargs[arg_key])
        elif (
            arg_key == "pme_threads_per_rank"
            or arg_key == "threads_per_pme_rank"
            or arg_key == "ntomp_pme"
        ):
            # TODO: Remove this temporary accommodation.
            assert not gmx.version.api_is_at_least(0, 2)
            if arg_key == "pme_threads_per_rank":
                warnings.warn(
                    "Key word pme_threads_per_rank has been renamed to threads_per_pme_rank.",
                    DeprecationWarning,
                )
            params["threads_per_pme_rank"] = int(kwargs[arg_key])
        elif arg_key == "steps" or arg_key == "nsteps":
            if kwargs[arg_key] is None:
                # None means "don't override the input" which is indicated by a parameter value of -2 in GROMACS 2019
                steps = -2
            else:
                # Otherwise we require steps to be a positive integer
                try:
                    steps = int(kwargs[arg_key])
                    if steps < 1:
                        raise exceptions.ValueError("steps to run must be at least 1")
                except (TypeError, ValueError) as e:
                    # steps is not an integer.
                    raise exceptions.TypeError(
                        '"steps" could not be interpreted as an integer.'
                    )
                # The "nsteps" command line flag will be removed in GROMACS 2020
                # and so "steps" is deprecated in gmxapi 0.0.7
                warnings.warn(
                    "`steps` keyword argument is deprecated. Consider `end_time` instead.",
                    DeprecationWarning,
                )
            params["steps"] = steps
        elif arg_key == "max_hours" or arg_key == "maxh":
            params["max_hours"] = float(kwargs[arg_key])
        elif arg_key == "append_output":
            # Try not to encourage confusion with the `mdrun` `-noappend` flag, which would be a confusing double negative if represented as a bool.
            params["append_output"] = bool(kwargs[arg_key])
        elif arg_key == "end_time":
            params[arg_key] = float(kwargs[arg_key])
        else:
            # Other arguments are accepted without checking or conversion.
            params[arg_key] = kwargs[arg_key]

    # Create an empty WorkSpec
    workspec = WorkSpec()

    # Create and add the Element for the tpr file(s)
    inputelement = WorkElement(
        namespace="gromacs", operation="load_tpr", params={"input": tpr_list}
    )
    inputelement.name = "tpr_input"
    if inputelement.name not in workspec.elements:
        # Operations such as this need to be replaced with accessors or properties that can check the validity of the WorkSpec
        workspec.elements[inputelement.name] = inputelement.serialize()
        inputelement.workspec = workspec

    # Create and add the simulation element
    # We can add smarter handling of the `depends` argument, but it is only critical to check when adding the element
    # to a WorkSpec.
    mdelement = WorkElement(operation="md", depends=[inputelement.name], params=params)
    mdelement.name = "md_sim"
    # Check that the element has not already been added, but that its dependency has.
    workspec.add_element(mdelement)

    return mdelement



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2022- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.
"""Test feature checking API and optional feature handling."""

from packaging.version import parse

try:
    from importlib.metadata import version, PackageNotFoundError
except ImportError:
    from importlib_metadata import version, PackageNotFoundError

import pytest
import gmxapi
import gmxapi._gmxapi as core
from gmxapi.utility import config


@pytest.mark.usefixtures("cleandir")
def test_feature_check():
    """Check the query features of the binary extension module."""
    assert hasattr(core, "has_feature")
    try:
        gmxpy_version = parse(version("gmxapi"))
    except PackageNotFoundError:
        # If the package distribution has not been built (such as when it has
        # merely been copied to a staging area, as in GROMACS builds through
        # at least 2022), the Python package metadata will not be available,
        # but through at least gmxapi 0.4, the `__version__` module attribute
        # is hard coded in the source files.
        gmxpy_version = parse(gmxapi.__version__)
    if gmxpy_version >= parse("0.4.0a3"):
        assert hasattr(core, "library_has_feature")
        assert "gmxapi_level" in config()
        assert parse(config()["gmxapi_level"]) >= parse("0.2")
        assert core.has_feature("create_context")
    assert not core.has_feature("spam")


@pytest.mark.skipif(
    not core.has_feature("mpi_bindings"),
    reason="Requires MPI bindings through mpi4py at package build time.",
)
@pytest.mark.withmpi_only
@pytest.mark.usefixtures("cleandir")
def test_mpi_bindings():
    from mpi4py import MPI

    rank_number = MPI.COMM_WORLD.Get_rank()
    comm_size = MPI.COMM_WORLD.Get_size()

    gmxapi_rank, gmxapi_size = core.mpi_report(MPI.COMM_WORLD)
    assert (gmxapi_rank, gmxapi_size) == (rank_number, comm_size)


@pytest.mark.skipif(
    not core.has_feature("mpi_bindings"),
    reason="Requires MPI bindings through mpi4py at package build time.",
)
@pytest.mark.withmpi_only
@pytest.mark.usefixtures("cleandir")
def test_mpi_sharing():
    from mpi4py import MPI

    rank_number = MPI.COMM_WORLD.Get_rank()

    # For non-MPI GROMACS, only offer a communicator of size 1.
    # We may be able to make resource allocation decisions based on larger
    # communicators in the future, but until then we should reject the
    # potentially wasteful resource assignment.
    color = 0
    if gmxapi.utility.config()["gmx_mpi_type"] != "library":
        if rank_number > 0:
            color = MPI.UNDEFINED
    sub_communicator = MPI.COMM_WORLD.Split(color, rank_number)
    if sub_communicator != MPI.COMM_NULL:
        try:
            api_context = core.create_context(sub_communicator)
            del api_context
        finally:
            sub_communicator.Free()

    # All GROMACS builds should be able to handle an ensemble of simulations
    # on disjoint sub-communicators of size 1.
    sub_communicator = MPI.COMM_WORLD.Split(rank_number, rank_number)
    if sub_communicator != MPI.COMM_NULL:
        try:
            api_context = core.create_context(sub_communicator)
            del api_context
        finally:
            sub_communicator.Free()

    # For MPI GROMACS, check that we can accept communicators of size > 1.
    if gmxapi.utility.config()["gmx_mpi_type"] == "library":
        ensemble_size = 2
        if rank_number < ensemble_size:
            color = 0
        else:
            color = MPI.UNDEFINED

        sub_communicator = MPI.COMM_WORLD.Split(color, rank_number)
        if sub_communicator != MPI.COMM_NULL:
            try:
                api_context = core.create_context(sub_communicator)
                del api_context
            finally:
                sub_communicator.Free()



#!/usr/bin/env python3
#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2014- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

import sys
import functools

from fnmatch import fnmatch

"""Central issue reporting implementation.

This module implements a Reporter class that is used by other Python modules in
this directory to report issues.  This allows central customization of the
output format, and also a central implementation for redirecting/copying
the output into a log file.  This class also implements sorting for the
messages such that all issues from a single file are reported next to each
other in the output, as well as filtering to make it possible to suppress
certain messages.
"""


@functools.total_ordering
class Location(object):
    """Location for a reported message."""

    def __init__(self, filename, line):
        """Create a location with the given file and line number.

        One or both of the parameters can be None, but filename should be
        specified if line is.
        """
        self.filename = filename
        self.line = line

    def __bool__(self):
        """Make empty locations False in boolean context."""
        return self.filename is not None

    def __str__(self):
        """Format the location as a string."""
        if self.line:
            return "{0}:{1}".format(self.filename, self.line)
        elif self.filename:
            return self.filename
        else:
            return "<unknown>"

    def __eq__(self, other):
        """Sort locations based on file name and line number."""
        return self.filename == other.filename and self.line == other.line

    def __lt__(self, other):
        """Sort locations based on file name and line number."""
        if self.filename != other.filename:
            if other.filename is None:
                return False
            if self.filename is None:
                return True
            return self.filename < other.filename
        else:
            if not self.filename:
                return False
            if other.line is None:
                return False
            if self.line is None:
                return True
            return self.line < other.line


@functools.total_ordering
class Message(object):
    """Single reported message.

    This class stores the contents of a reporter message for later output to
    allow sorting the output messages reasonably by the reported location.
    """

    def __init__(self, message, details=None, filename=None, location=None):
        """Create a message object.

        The message parameter provides the actual text, while optional details
        provides a list of extra lines that provide context information for the
        error.  filename and location provide two alternative ways of
        specifying the location of the issue:
         - if filename is provided, the issue is reported in that file, without
           a line number
         - if location is provided, it should be a Location instance
        """
        if filename:
            self.location = Location(filename, None)
        elif location:
            self.location = location
        else:
            self.location = Location(None, None)
        self.message = message
        self.details = details

    def __eq__(self, other):
        """Sort messages based on file name and line number."""
        return self.location == other.location

    def __lt__(self, other):
        """Sort messages based on file name and line number."""
        return self.location < other.location


class Filter(object):
    """Filter expression to exclude messages."""

    def __init__(self, filterline):
        """Initialize a filter from a line in a filter file."""
        self._orgline = filterline
        filepattern, text = filterline.split(":", 1)
        if filepattern == "*":
            self._filematcher = lambda x: x is not None
        elif filepattern:
            self._filematcher = lambda x: x and fnmatch(x, "*/" + filepattern)
        else:
            self._filematcher = lambda x: x is None
        self._textpattern = text.strip()
        self._count = 0

    def matches(self, message):
        """Check whether the filter matches a message."""
        if not self._filematcher(message.location.filename):
            return False
        if not fnmatch(message.message, self._textpattern):
            return False
        self._count += 1
        return True

    def get_match_count(self):
        """Return the number of times this filter has matched."""
        return self._count

    def get_text(self):
        """Return original line used to specify the filter."""
        return self._orgline


class Reporter(object):
    """Collect and write out issues found by checker scripts."""

    def __init__(self, logfile=None, quiet=False):
        """Initialize the reporter.

        If logfile is set to a file name, all issues will be written to this
        file in addition to stderr.

        If quiet is set to True, the reporter will suppress all output.
        """
        self._logfp = None
        if logfile:
            self._logfp = open(logfile, "w")
        self._messages = []
        self._filters = []
        self._quiet = quiet
        self._had_warnings = False

    def _write(self, message):
        """Implement actual message writing."""
        wholemsg = ""
        if message.location:
            wholemsg += str(message.location) + ": "
        wholemsg += message.message
        if message.details:
            wholemsg += "\n    " + "\n    ".join(message.details)
        wholemsg += "\n"
        sys.stderr.write(wholemsg)
        if self._logfp:
            self._logfp.write(wholemsg)
        self._had_warnings = True

    def _report(self, message):
        """Handle a single reporter message."""
        if self._quiet:
            return
        for filterobj in self._filters:
            if filterobj.matches(message):
                return
        if not message.location:
            self._write(message)
        else:
            self._messages.append(message)

    def load_filters(self, filterfile):
        """Load filters for excluding messages from a file."""
        with open(filterfile, "r") as fp:
            for filterline in fp:
                filterline = filterline.strip()
                if not filterline or filterline.startswith("#"):
                    continue
                self._filters.append(Filter(filterline))

    def write_pending(self):
        """Write out pending messages in sorted order."""
        self._messages.sort()
        for message in self._messages:
            self._write(message)
        self._messages = []

    def report_unused_filters(self):
        """Report filters that did not match any messages."""
        for filterobj in self._filters:
            if filterobj.get_match_count() == 0:
                # TODO: Consider adding the input filter file as location
                text = "warning: unused filter: " + filterobj.get_text()
                self._write(Message(text))

    def had_warnings(self):
        """Return true if any warnings have been reported."""
        return self._had_warnings

    def close_log(self):
        """Close the log file if one exists."""
        assert not self._messages
        if self._logfp:
            self._logfp.close()
            self._logfp = None

    def xml_assert(self, xmlpath, message):
        """Report issues in Doxygen XML that violate assumptions in the script."""
        self._report(Message("warning: " + message, filename=xmlpath))

    def input_error(self, message):
        """Report issues in input files."""
        self._report(Message("error: " + message))

    def file_error(self, fileobj, message):
        """Report file-level issues."""
        self._report(
            Message("error: " + message, location=fileobj.get_reporter_location())
        )

    def code_issue(self, entity, message, details=None):
        """Report an issue in a code construct (not documentation related)."""
        self._report(
            Message(
                "warning: " + message, details, location=entity.get_reporter_location()
            )
        )

    def cyclic_issue(self, message, details=None):
        """Report a cyclic dependency issue."""
        self._report(Message("warning: " + message, details))

    def doc_error(self, entity, message):
        """Report an issue in documentation."""
        self._report(
            Message(
                "error: " + entity.get_name() + ": " + message,
                location=entity.get_reporter_location(),
            )
        )

    def doc_note(self, entity, message):
        """Report a potential issue in documentation."""
        self._report(
            Message(
                "note: " + entity.get_name() + ": " + message,
                location=entity.get_reporter_location(),
            )
        )



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2019- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""Tests the interfaces defined in operation.py and behavior of simple operations.

See also test_commandline.py for integration tests.
"""

import os
import shutil
import stat
import tempfile
from typing import NamedTuple

import gmxapi as gmx
from gmxapi import commandline_operation
from gmxapi.operation import ResultDescription


def test_comparison():
    int_array_description = ResultDescription(dtype=int, width=3)
    same_description = ResultDescription(dtype=int, width=3)
    different_type_description = ResultDescription(dtype=float, width=3)
    different_width_description = ResultDescription(dtype=int, width=1)
    assert int_array_description == same_description
    assert int_array_description != different_width_description
    assert int_array_description != different_type_description

    class CompatibleRepresentation(NamedTuple):
        dtype: type
        width: int

    equivalent_description = CompatibleRepresentation(dtype=int, width=3)
    assert int_array_description == equivalent_description
    assert equivalent_description == int_array_description
    assert equivalent_description != different_type_description
    assert different_type_description != equivalent_description
    assert equivalent_description != different_width_description
    assert different_width_description != equivalent_description


def test_scalar():
    operation = gmx.make_constant(42)
    assert isinstance(operation.dtype, type)
    assert operation.dtype == int
    assert operation.result() == 42


def test_list():
    list_a = [1, 2, 3]

    # TODO: test input validation
    list_result = gmx.concatenate_lists(sublists=[list_a])
    assert list_result.dtype == gmx.datamodel.NDArray
    # Note: this is specifically for the built-in tuple type.
    # Equality comparison may work differently for different sequence types.
    assert tuple(list_result.result()) == tuple(list_a)
    assert len(list_result.result()) == len(list_a)

    list_result = gmx.concatenate_lists([list_a, list_a])
    assert len(list_result.result()) == len(list_a) * 2
    assert tuple(list_result.result()) == tuple(list_a + list_a)

    list_b = gmx.ndarray([42])

    list_result = gmx.concatenate_lists(sublists=[list_b])
    assert list_result.result()[0] == 42

    list_result = gmx.join_arrays(front=list_a, back=list_b)
    assert len(list_result.result()) == len(list_a) + 1
    assert tuple(list_result.result()) == tuple(list(list_a) + [42])


def test_data_dependence(cleandir):
    """Test dependent sequence of operations.

    Confirm that data dependencies correctly establish resolvable execution dependencies.

    In a sequence of two operations, write a two-line file one line at a time.
    Use the output of one operation as the input of another.
    """
    with tempfile.TemporaryDirectory() as directory:
        file1 = os.path.join(directory, "input")
        file2 = os.path.join(directory, "output")

        # Make a shell script that acts like the type of tool we are wrapping.
        scriptname = os.path.join(directory, "clicommand.sh")
        with open(scriptname, "w") as fh:
            fh.write(
                "\n".join(
                    [
                        "#!" + shutil.which("bash"),
                        "# Concatenate an input file and a string argument to an output file.",
                        "# Mock a utility with the tested syntax.",
                        '#     clicommand.sh "some words" -i inputfile -o outputfile',
                        "echo $1 | cat $3 - > $5\n",
                    ]
                )
            )
        os.chmod(scriptname, stat.S_IRWXU)

        line1 = "first line"
        filewriter1 = commandline_operation(
            scriptname,
            arguments=[line1],
            input_files={"-i": os.devnull},
            output_files={"-o": file1},
        )

        line2 = "second line"
        filewriter2 = commandline_operation(
            scriptname,
            arguments=[line2],
            input_files={"-i": filewriter1.output.file["-o"]},
            output_files={"-o": file2},
        )

        filewriter2.run()
        # Check that the files have the expected lines
        with open(filewriter1.output.file["-o"].result(), "r") as fh:
            lines = [text.rstrip() for text in fh]
        assert len(lines) == 1
        assert lines[0] == line1
        with open(filewriter2.output.file["-o"].result(), "r") as fh:
            lines = [text.rstrip() for text in fh]
        assert len(lines) == 2
        assert lines[0] == line1
        assert lines[1] == line2

        try:
            # If we're running with MPI, check that the file is accessible
            # on all ranks, but make sure that we don't delete the temporary
            # directory before all ranks have looked for it.
            from mpi4py import MPI

            MPI.COMM_WORLD.barrier()
        except ImportError:
            pass
